// =============================================================================
//
//   Cel-Shading Demo
//   glSlang-Port
//
// =============================================================================
//   Copyright  2003-2004 by Sascha Willems - http://www.delphigl.de
//   visit the Delphi OpenGL Community - http://www.delphigl.com
// =============================================================================
//   Contents of this file are subject to the GNU Public License (GPL) which can
//   be obtained here : http://opensource.org/licenses/gpl-license.php
//   So only use this file if you fully unterstand that license!!!
// =============================================================================
//   Your graphicsadaptor must support the following extensions to run this :
//    - GL_ARB_Shading_Language_100
//    - GL_ARB_Shader_Objects
//    - GL_ARB_Vertex_Shader
//    - GL_ARB_Fragment_Shader
//    - WGL_ARB_Pbuffer
// =============================================================================
//   Short description :
//    This demo renders a celshaded scene (celshaded using celShade.vert and
//    celShade.frag) in two passes to a texture (Pixelbuffer), and then draws
//    this texture to screen using a postscene-effect on that texture.
// =============================================================================

unit CelShade_MainForm;

interface

uses
  Windows,
  SysUtils,
  Classes,
  Graphics,
  Controls,
  Forms,
  Dialogs,
  AppEvnts,
  Math,
  ExtCtrls,

  glFont,
  gl3DS,
  glPBuffer,
  dglOpenGL,
  Textures;

type
  TGLForm = class(TForm)
    ApplicationEvents : TApplicationEvents;
    FPSTimer          : TTimer;
    RippleTimer       : TTimer;
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure ApplicationEventsIdle(Sender: TObject; var Done: Boolean);
    procedure FormMouseMove(Sender: TObject; Shift: TShiftState; X, Y: Integer);
    procedure FPSTimerTimer(Sender: TObject);
    procedure FormMouseDown(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
    procedure FormMouseUp(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
    procedure RippleTimerTimer(Sender: TObject);
    procedure FormKeyDown(Sender: TObject; var Key: Word; Shift: TShiftState);
  public
    Direction     : array[0..1] of Single;
    Position      : array[0..2] of Single;
    Mesh3DS       : TAll3DSMesh;
    TerrainList   : glUInt;
    SkyDomeList   : glUInt;
    CelTexture    : glUInt;
    SkyTexture    : glUInt;
    OldMPos       : TPoint;
    FontBase      : glUInt;
    RC            : HGLRC;
    DC            : HDC;
    Frame         : Word;
    Move          : ShortInt;
    FPS           : Integer;
    procedure ShowInfo;
    procedure DrawScene;
    procedure RenderPass(pPassNr : Byte);
    procedure LoadLinkAndCompileShaders(var pProgramObject : GLHandleARB; pVertexShaderFile, pFragmentShaderFile : String);
  end;

var
 GLForm               : TGLForm;

 VertexShaderObject   : GLHandleARB;
 FragmentShaderObject : GLHandleARB;
 ProgramObject        : GLHandleARB;
 ProgramObjectRipple  : GLHandleARB;
 ProgramObjectFishEye : GLHandleARB;

 RippleTime           : Single;
 RippleWaveFreq       : Single = 50;
 RippleWaveAmp        : Single = 0.012;
 FishEyeLensRadius    : Single = 1;
 FishEyeSignCurvature : Single = -15;

 PostSceneShader      : Byte;

 ShowInfoBlock        : Boolean;
 DoRender             : Boolean;

implementation

{$R *.dfm}

// =============================================================================
//  DrawHalfSphere
// =============================================================================
procedure DrawHalfSphere(cX, cY, cZ, Radius : TGLFloat; n : Integer);
var
 i,j                  : Integer;
 theta1,theta2,theta3 : TGLFloat;
 X, Y, Z, px, py, pz  : TGLFloat;
begin
for j := 0 to n div 4 -1 do
 begin
 theta1 := J*2*PI/N - PI/2;
 theta2 := (J+1)*2*PI/n - PI/2;
 glBegin(GL_TRIANGLE_STRIP);
  for i := 0 to n do
   begin
   theta3 := i*2*PI/N;
   x := cos(theta2) * cos(theta3);
   y := sin(theta2);
   z := cos(theta2) * sin(theta3);
   px := CX + Radius*x;
   py := CY + Radius*y;
   pz := CZ + Radius*z;
   glNormal3f(X,Y,Z);
   glTexCoord2f(-1-I/n, 2*(J+1)/n);
   glVertex3f(px,py,pz);
   X := cos(theta1) * cos(theta3);
   Y := sin(theta1);
   Z := cos(theta1) * sin(theta3);
   px := CX + Radius*X;
   py := CY + Radius*Y;
   pz := CZ + Radius*Z;
   glNormal3f(X,Y,Z);
   glTexCoord2f(-1-i/n, 2*j/n);
   glVertex3f(px,py,pz);
   end;
 glEnd;
 end;
end;

// =============================================================================
//  TGLForm.LoadLinkAndCompileShaders
// =============================================================================
procedure TGLForm.LoadLinkAndCompileShaders(var pProgramObject : GLHandleARB; pVertexShaderFile, pFragmentShaderFile : String);
var
 VShader : GLHandleARB;
 FShader : GLHandleARB;
 SSource : TStringList;
 Source  : String;
 SLen    : Integer;
begin
pProgramObject := glCreateProgramObjectARB;

VShader        := glCreateShaderObjectARB(GL_VERTEX_SHADER_ARB);
FShader        := glCreateShaderObjectARB(GL_FRAGMENT_SHADER_ARB);

SSource        := TStringList.Create;

SSource.LoadFromFile(pVertexShaderFile);
Source := SSource.Text;
SLen   := Length(Source);
glShaderSourceARB(VShader, 1, @Source, @SLen);

SSource.LoadFromFile(pFragmentShaderFile);
Source := SSource.Text;
SLen   := Length(Source);
glShaderSourceARB(FShader, 1, @Source, @SLen);
SSource.Free;

glCompileShaderARB(VShader);
glCompileShaderARB(FShader);

glAttachObjectARB(pProgramObject, VShader);
glAttachObjectARB(pProgramObject, FShader);

glDeleteObjectARB(VShader);
glDeleteObjectARB(FShader);

glLinkProgramARB(pProgramObject);
end;

// =============================================================================
//  TGLForm.FormCreate
// =============================================================================
procedure TGLForm.FormCreate(Sender: TObject);
begin
if MessageDlg('Fullscreen?', mtConfirmation, [mbYes, mbNo], 0) = mrYes then
 begin
 BorderStyle := bsNone;
 WindowState := wsMaximized;
 Cursor      := crNone;
 end;
ChDir(ExtractFilePath(Application.ExeName));
InitOpenGL;
DC := GetDC(Handle);
RC := CreateRenderingContext(DC, [opDoubleBuffered], 32, 0, 0, 0, 0, 0);
ActivateRenderingContext(DC, RC);
TextureFont := TTexFont.Create('texture\font.tga');

// Create pixelbuffer and load celshading shader
PixelBuffer := TPixelBuffer.Create(1024, 1024, DC, RC);
PixelBuffer.Enable;
// Share lists and textures on RCs
wglShareLists(RC, PixelBuffer.RC);
glEnable(GL_DEPTH_TEST);
glDepthFunc(GL_LEQUAL);
glEnable(GL_TEXTURE_1D);
glClearColor(0.2,0.4,0.7,1);

// Load textures
Textures.Target := GL_TEXTURE_1D;
LoadTexture('texture\celshade.tga', CelTexture);
Textures.Target := GL_TEXTURE_2D;
LoadTexture('texture\skydome.tga', SkyTexture);
// Generate displaylist for the skydome
SkyDomeList := glGenLists(1);
glNewList(SkyDomeList, GL_COMPILE);
 glColor3f(1,1,1);
 glDisable(GL_CULL_FACE);
 glDepthMask(False);
 glActiveTexture(GL_TEXTURE1);
 glDisable(GL_TEXTURE_2D);
 glActiveTexture(GL_TEXTURE0);
 glEnable(GL_TEXTURE_2D);
 glBindTexture(GL_TEXTURE_2D, SkyTexture);
 glScalef(1,-1,1);
 DrawHalfSphere(0,0.4,0, 1, 32);
 glScalef(1,-1,1);
 glDepthMask(True);
glEndList;
// Load terrain from 3DS-File
Mesh3DS := TAll3DSMesh.Create(self);
Mesh3DS.TexturePath := 'texture\';
Mesh3DS.LoadFromFile('models\terrain.3ds');
// Generate displaylist for the terrain
TerrainList := glGenLists(1);
glNewList(TerrainList, GL_COMPILE);
 glPushMatrix;
  glScalef(0.1, 0.1, 0.1);
  glRotatef(-90, 1,0,0);
  glActiveTexture(GL_TEXTURE1);
  glBindTexture(GL_TEXTURE_1D, CelTexture);
  glEnable(GL_TEXTURE_1D);
  glActiveTexture(GL_TEXTURE0);
  Mesh3DS.Render;
  glActiveTexture(GL_TEXTURE1);
  glDisable(GL_TEXTURE_1D);
 glPopMatrix;
glEndList;
glActiveTexture(GL_TEXTURE0);

LoadLinkAndCompileShaders(ProgramObject, 'glSlangShader\celShade.vert', 'glSlangShader\celShade.frag');
glUseProgramObjectARB(ProgramObject);
glUniform4fARB(glGetUniformLocationARB(ProgramObject, PGLCharARB('LightPosition')), 512,25,30,0);
glUniform4fARB(glGetUniformLocationARB(ProgramObject, PGLCharARB('CelColor')), 1,1,1,0);
glUniform1iARB(glGetUniformLocationARB(ProgramObject, PGLCharARB('tex0')), 0);
glUniform1iARB(glGetUniformLocationARB(ProgramObject, PGLCharARB('tex1')), 1);

PixelBuffer.Disable;

// Fill our pixelbuffer's texture so that we can subcopy later on
glBindTexture(GL_TEXTURE_2D, PixelBuffer.TextureID);
glCopyTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, 0, 0, PixelBuffer.Width, PixelBuffer.Height, 0);

// Load rippleshader in normal context
LoadLinkAndCompileShaders(ProgramObjectFishEye, 'glSlangShader\ripple.vert', 'glSlangShader\fisheye.frag');
LoadLinkAndCompileShaders(ProgramObjectRipple, 'glSlangShader\ripple.vert', 'glSlangShader\ripple.frag');
glUseProgramObjectARB(0);


OldMPos := Mouse.CursorPos;
Position[1] := 2.5;

ShowInfoBlock := True;
DoRender      := True;
end;

// =============================================================================
//  TGLForm.FormDestroy
// =============================================================================
procedure TGLForm.FormDestroy(Sender: TObject);
begin
TextureFont.Free;
Mesh3DS.Free;
PixelBuffer.Free;
DeActivateRenderingContext;
wglDeleteContext(RC);
ReleaseDC(Handle, DC);
end;

// =============================================================================
//  TGLForm.RenderPass
// =============================================================================
procedure TGLForm.RenderPass(pPassNr : Byte);
begin
// Pass 1 = Cel-Shading
if pPassNr = 1 then
 begin
 glUniform1iARB(glGetUniformLocationARB(ProgramObject, PChar('pass')), 1);
 glEnable(GL_CULL_FACE);
 glCullFace(GL_FRONT);
 glCallList(TerrainList);
 end;
// Pass 2 = Outlines
if pPassNr = 2 then
 begin
 glUniform1iARB(glGetUniformLocationARB(ProgramObject, PChar('pass')), 2);
 glDisable(GL_TEXTURE_1D);
 glEnable(GL_CULL_FACE);
 glCullFace(GL_BACK);
 glLineWidth(10);
 glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
 glCallList(TerrainList);
 glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
 end;
end;

// =============================================================================
//  TGLForm.ShowInfo
// =============================================================================
procedure TGLForm.ShowInfo;
begin
glMatrixMode(GL_PROJECTION);
glLoadIdentity;
glOrtho(0, 640, 480, 0, -1, 1);
glMatrixMode(GL_MODELVIEW);
glLoadIdentity;
glActiveTexture(GL_TEXTURE1);
glDisable(GL_TEXTURE_2D);
glActiveTexture(GL_TEXTURE0);
glDisable(GL_TEXTURE_2D);

glEnable(GL_BLEND);
glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

glColor4f(0,0,0, 0.8);
glBegin(GL_QUADS);
 glVertex2f(0,   0);
 glVertex2f(250, 0);
 glVertex2f(250, 100);
 glVertex2f(0,   100);
glEnd;
glDisable(GL_BLEND);

glEnable(GL_TEXTURE_2D);
glEnable(GL_ALPHA_TEST);
glAlphaFunc(GL_GREATER, 0.1);
TextureFont.Print2D('FPS : '+IntToStr(FPS), [4,0,0], [1,1,1], 1, 1);

case PostSceneShader of
 0 : begin
     TextureFont.Print2D('Postscene shader : ripple (F5)', [4,12,0], [1,1,1], 1, 1);
     TextureFont.Print2D('Shaderparameters :', [4,30,0], [1,1,1], 1, 1);
     TextureFont.Print2D('WaveFreq = '+FloatToStrF(RippleWaveFreq, ffFixed, 6, 4)+' [F1/F2]', [8,44,0], [1,1,1], 1, 1);
     TextureFont.Print2D('WaveAmp = '+FloatToStrF(RippleWaveAmp, ffFixed, 6, 4)+' [F3/F4]', [8,56,0], [1,1,1], 1, 1);
     end;
 1 : begin
     TextureFont.Print2D('Postscene shader : fisheye (F5)', [4,12,0], [1,1,1], 1, 1);
     TextureFont.Print2D('Shaderparameters :', [4,30,0], [1,1,1], 1, 1);
     TextureFont.Print2D('SignCurvature = '+FloatToStrF(FishEyeSignCurvature, ffFixed, 6, 4)+' [F1/F2]', [8,44,0], [1,1,1], 1, 1);
     TextureFont.Print2D('LensRadius = '+FloatToStrF(FishEyeLensRadius, ffFixed, 6, 4)+' [F3/F4]', [8,56,0], [1,1,1], 1, 1);
     end;
 2 : TextureFont.Print2D('Postscene shader : none (F5)', [4,12,0], [1,1,1], 1, 1);
end;

TextureFont.Print2D('F6 - Toggle this info', [4,84,0], [1,1,1], 1, 1);
glDisable(GL_ALPHA_TEST);
end;

// =============================================================================
//  TGLForm.DrawScene
// =============================================================================
procedure TGLForm.DrawScene;
begin
// Skydome (with shaders disabled)
glUseProgramObjectARB(0);
glCallList(SkyDomeList);
// Terrain (with celshader enabled)
glUseProgramObjectARB(ProgramObject);
glTranslatef(-Position[0], -Position[1], -Position[2]);
GLForm.RenderPass(1);
GLForm.RenderPass(2);
end;

// =============================================================================
//  TGLForm.ApplicationEventsIdle
// =============================================================================
procedure TGLForm.ApplicationEventsIdle(Sender: TObject; var Done: Boolean);
begin
if not DoRender then
 exit;
// Render the whole Scene to our Pixelbuffer ===================================
PixelBuffer.Enable;
glUseProgramObjectARB(ProgramObject);

glMatrixMode(GL_PROJECTION);
glLoadIdentity;
glViewport(0, 0, PixelBuffer.Width, PixelBuffer.Height);
gluPerspective(90, ClientWidth/ClientHeight, 0.1, 2048);
glMatrixMode(GL_MODELVIEW);
glLoadIdentity;
glClear(GL_COLOR_BUFFER_BIT or GL_DEPTH_BUFFER_BIT);

glRotatef(Direction[0], 1,0,0);
glRotatef(Direction[1], 0,1,0);

DrawScene;

PixelBuffer.Disable;

// Send parameters to our post scene shader ====================================
if PostSceneShader = 0 then
 begin
 glUseProgramObjectARB(ProgramObjectRipple);
 glUniform1fARB(glGetUniformLocationARB(ProgramObjectRipple, PGLCharARB('waveFreq')), RippleWaveFreq);
 glUniform1fARB(glGetUniformLocationARB(ProgramObjectRipple, PGLCharARB('waveAmp')), RippleWaveAmp);
 glUniform1fARB(glGetUniformLocationARB(ProgramObjectRipple, PGLCharARB('time')), RippleTime);
 end;
if PostSceneShader = 1 then
 begin
 glUseProgramObjectARB(ProgramObjectFishEye);
 glUniform1fARB(glGetUniformLocationARB(ProgramObjectFishEye, PGLCharARB('signcurvature')), FishEyeSignCurvature);
 glUniform1fARB(glGetUniformLocationARB(ProgramObjectFishEye, PGLCharARB('lensradius')), FishEyeLensRadius);
 glUniform1fARB(glGetUniformLocationARB(ProgramObjectFishEye, PGLCharARB('inverse')), 0);
 end;
if PostSceneShader = 2 then
 glUseProgramObjectARB(0);

// Now render the scenetexture to screen =======================================
glMatrixMode(GL_PROJECTION);
glLoadIdentity;
glViewport(0, 0, ClientWidth, ClientHeight);
glMatrixMode(GL_MODELVIEW);
glLoadIdentity;

glEnable(GL_TEXTURE_2D);
PixelBuffer.Bind;
glBegin(GL_QUADS);
 glTexCoord2f(0,0); glVertex3f(-1, -1, 0);
 glTexCoord2f(0,1); glVertex3f(-1,  1, 0);
 glTexCoord2f(1,1); glVertex3f( 1,  1, 0);
 glTexCoord2f(1,0); glVertex3f( 1, -1, 0);
glEnd;
PixelBuffer.Release;

glUseProgramObjectARB(0);

if ShowInfoBlock then
 ShowInfo;

SwapBuffers(DC);

Done := False;
inc(Frame);

Position[0] := Position[0]+0.1*Sin(DegToRad(Direction[1]))*Move;
Position[1] := Position[1]-0.1*Sin(DegToRad(Direction[0]))*Move;
Position[2] := Position[2]-0.1*Cos(DegToRad(Direction[1]))*Move;
end;

// =============================================================================
//  TGLForm.FormMouseMove
// =============================================================================
procedure TGLForm.FormMouseMove(Sender: TObject; Shift: TShiftState; X,Y: Integer);
var
 MPos : TPoint;
begin
MPos := Mouse.CursorPos;
Direction[0] := Direction[0]+(MPos.Y-OldMPos.Y)*0.25;
Direction[1] := Direction[1]+(MPos.X-OldMPos.X)*0.25;
OldMPos := MPos;
if MPos.X < Left+15 then
 SetCursorPos(Left+ClientWidth-17, MPos.Y);
if MPos.X > Left+ClientWidth-15 then
 SetCursorPos(Left+17, MPos.Y);
if MPos.Y < Top+35 then
 SetCursorPos(MPos.X, Top+ClientHeight-17);
if MPos.Y > Top+ClientHeight-15 then
 SetCursorPos(MPos.X, Top+37);
OldMPos := Mouse.CursorPos;
end;

// =============================================================================
//  TGLForm.FPSTimerTimer
// =============================================================================
procedure TGLForm.FPSTimerTimer(Sender: TObject);
begin
FPS     := Frame;
Frame   := 0;
end;

// =============================================================================
//  TGLForm.FormMouseDown
// =============================================================================
procedure TGLForm.FormMouseDown(Sender: TObject; Button: TMouseButton;Shift: TShiftState; X, Y: Integer);
begin
case Button of
 mbLeft  : Move := 1;
 mbRight : Move := -1;
end;
end;

// =============================================================================
//  TGLForm.FormMouseUp
// =============================================================================
procedure TGLForm.FormMouseUp(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
Move := 0;
end;

// =============================================================================
//  TGLForm.RippleTimerTimer
// =============================================================================
procedure TGLForm.RippleTimerTimer(Sender: TObject);
begin
RippleTime := RippleTime+0.01;
if RippleTime > 1000 then
 RippleTime := 0;
end;

// =============================================================================
//  TGLForm.FormKeyDown
// =============================================================================
procedure TGLForm.FormKeyDown(Sender: TObject; var Key: Word; Shift: TShiftState);
begin
case PostSceneShader of
 0 :  case Key of
       vk_F1 : RippleWaveFreq := RippleWaveFreq-0.5;
       vk_F2 : RippleWaveFreq := RippleWaveFreq+0.5;
       vk_F3 : RippleWaveAmp  := RippleWaveAmp-0.001;
       vk_F4 : RippleWaveAmp  := RippleWaveAmp+0.001;
      end;
 1 :  case Key of
       vk_F1 : FishEyeSignCurvature := FishEyeSignCurvature-0.5;
       vk_F2 : FishEyeSignCurvature := FishEyeSignCurvature+0.5;
       vk_F3 : FishEyeLensRadius  := FishEyeLensRadius-0.01;
       vk_F4 : FishEyeLensRadius  := FishEyeLensRadius+0.01;
      end;
end;
case Key of
 vk_F5 : begin
         inc(PostSceneShader);
         if PostSceneShader > 2 then
          PostSceneShader := 0;
         end;
 vk_F6 : ShowInfoBlock := not ShowInfoBlock;
end;
end;

end.
