// =============================================================================
//
//   Cel-Shading Demo
//
// =============================================================================
//   Copyright  2003 by Sascha Willems - http://www.delphigl.de
// =============================================================================
//  Contents of this file are subject to the GNU Public License (GPL) which can
//  be obtained here : http://opensource.org/licenses/gpl-license.php
// =============================================================================
//  V2 : Added a neat Skydome to give some atmosphere ;)
// =============================================================================

unit CelShade_MainForm;

interface

uses
  Windows,
  Messages,
  SysUtils,
  Variants,
  Classes,
  Graphics,
  Controls,
  Forms,
  Dialogs,
  AppEvnts,
  Math,
  ExtCtrls,

  gl3DS,
  dglOpenGL,
  Textures,
  fxPascal;

type
  TGLForm = class(TForm)
    ApplicationEvents : TApplicationEvents;
    FPSTimer: TTimer;
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure ApplicationEventsIdle(Sender: TObject; var Done: Boolean);
    procedure FormMouseMove(Sender: TObject; Shift: TShiftState; X, Y: Integer);
    procedure FPSTimerTimer(Sender: TObject);
    procedure FormMouseDown(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
    procedure FormMouseUp(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
    procedure FormKeyPress(Sender: TObject; var Key: Char);
  public
    Direction   : array[0..1] of Single;
    Position    : array[0..2] of Single;
    VShader     : array[0..1] of TGLUInt;
    Mesh3DS     : TAll3DSMesh;
    TerrainList : TGLUInt;
    SkyDomeList : TGLUInt;
    CelTexture  : TGLUInt;
    SkyTexture  : TGLUInt;
    OldMPos     : TPoint;
    RC          : HGLRC;
    DC          : HDC;
    Frame       : Word;
    Move        : ShortInt;
    procedure LoadVertexShader;
    procedure RenderPass(pPassNr : Byte);
  end;

var
  GLForm : TGLForm;

implementation

{$R *.dfm}

// =============================================================================
//  DrawHalfSphere
// =============================================================================
procedure DrawHalfSphere(cX, cY, cZ, Radius : TGLFloat; n : Integer);
var
 i,j                  : Integer;
 theta1,theta2,theta3 : TGLFloat;
 X, Y, Z, px, py, pz  : TGLFloat;
begin
if Radius < 0 then
 Radius :=-Radius;
if n < 0 then
 n := -n;
if (n < 4) or (Radius <= 0) then
 begin
 glBegin(GL_POINTS);
  glVertex3f(CX, CY, CZ);
 glEnd;
 exit;
 end;
for j := 0 to n div 4 -1 do
 begin
 theta1 := J*2*PI/N - PI/2;
 theta2 := (J+1)*2*PI/n - PI/2;
 glBegin(GL_TRIANGLE_STRIP);
 for i := 0 to n do
  begin
  theta3 := i*2*PI/N;
  x := cos(theta2) * cos(theta3);
  y := sin(theta2);
  z := cos(theta2) * sin(theta3);
  px := CX + Radius*x;
  py := CY + Radius*y;
  pz := CZ + Radius*z;
  glNormal3f(X,Y,Z);
  glTexCoord2f(-1-I/n, 2*(J+1)/n);
  glVertex3f(px,py,pz);
  X := cos(theta1) * cos(theta3);
  Y := sin(theta1);
  Z := cos(theta1) * sin(theta3);
  px := CX + Radius*X;
  py := CY + Radius*Y;
  pz := CZ + Radius*Z;
  glNormal3f(X,Y,Z);
  glTexCoord2f(-1-i/n, 2*j/n);
  glVertex3f(px,py,pz);
  end;
 glEnd;
 end;
end;

// =============================================================================
//  TGLForm.LoadVertexShader
// =============================================================================
procedure TGLForm.LoadVertexShader;
const
 VShaderFile : array[0..1] of String = ('vshader_pass1.txt', 'vshader_pass2.txt');
var
 i          : Integer;
 ARBProgram : String;
 ProgramStr : TStringList;
begin
for i := 0 to High(VShaderFile) do
 begin
 // ARB-Programm aus fxPascal-Shader kompilieren
 ProgramStr := TStringList.Create;
 ProgramStr.LoadFromFile('shader\'+VShaderFile[i]);
 ARBProgram := CompileProgram(ProgramStr.Text);
 ProgramStr.Free;
 // Vertexprogramm generieren
 glGenProgramsARB(1, @VShader[i]);
 glBindProgramARB(GL_VERTEX_PROGRAM_ARB, VShader[i]);
 glProgramStringARB(GL_VERTEX_PROGRAM_ARB, GL_PROGRAM_FORMAT_ASCII_ARB, Length(ARBProgram), PChar(ARBProgram));
 end;
end;

// =============================================================================
//  TGLForm.FormCreate
// =============================================================================
procedure TGLForm.FormCreate(Sender: TObject);
begin
ChDir(ExtractFilePath(Application.ExeName));
InitOpenGL;
DC := GetDC(Handle);
RC := CreateRenderingContext(DC, [opDoubleBuffered], 32, 24, 0, 0, 0, 0);
ActivateRenderingContext(DC, RC);
glEnable(GL_DEPTH_TEST);
glDepthFunc(GL_LEQUAL);
glClearColor(0.2,0.4,0.7,1);
// Texturen laden
LoadTexture('texture\celshade.tga', CelTexture, False, GL_LINEAR, GL_LINEAR_MIPMAP_LINEAR, False);
LoadTexture('texture\skydome.tga', SkyTexture, False, GL_LINEAR, GL_LINEAR_MIPMAP_LINEAR, False);
// 3DS-Modelle laden
Mesh3DS := ReadMesh3DS('models\terrain.3ds');
// Shader laden
LoadVertexShader;
// Displayliste erstellen
TerrainList := glGenLists(1);
glNewList(TerrainList, GL_COMPILE);
 glPushMatrix;
  glScalef(0.01,0.01,0.01);
  Render3ds(Mesh3DS);
 glPopMatrix;
glEndList;
OldMPos := Mouse.CursorPos;
Position[1] := 0.25;
// Skydomeliste
SkyDomeList := glGenLists(1);
glNewList(SkyDomeList, GL_COMPILE);
 glColor3f(1,1,1);
 glDisable(GL_CULL_FACE);
 glDisable(GL_VERTEX_PROGRAM_ARB);
 glDepthMask(False);
 glEnable(GL_TEXTURE_2D);
 glBindTexture(GL_TEXTURE_2D, SkyTexture);
 glScalef(1,-1,1);
 DrawHalfSphere(0,0,0, 1, 32);
 glScalef(1,-1,1);
 glDepthMask(True);
glEndList;
end;

// =============================================================================
//  TGLForm.FormDestroy
// =============================================================================
procedure TGLForm.FormDestroy(Sender: TObject);
begin
DeActivateRenderingContext;
wglDeleteContext(RC);
ReleaseDC(Handle, DC);
end;

// =============================================================================
//  TGLForm.RenderPass
// =============================================================================
procedure TGLForm.RenderPass(pPassNr : Byte);
begin
// Pass 1 = Cel-Shading
if pPassNr = 1 then
 begin
 glEnable(GL_CULL_FACE);
 glCullFace(GL_FRONT);
 // 1D-Textur binden
 glDisable(GL_TEXTURE_2D);
 glEnable(GL_TEXTURE_1D);
 glBindTexture(GL_TEXTURE_1D, CelTexture);
 // Vertexshader
 glEnable(GL_VERTEX_PROGRAM_ARB);
 glBindProgramARB(GL_VERTEX_PROGRAM_ARB, VShader[0]);
 // Programmparameter
 glProgramLocalParameter4fARB(GL_VERTEX_PROGRAM_ARB, 0, 0,200,0,0); // LightPos
 glProgramLocalParameter4fARB(GL_VERTEX_PROGRAM_ARB, 1, 1,1,0.6,0); // CelColor
 // Terrain
 glCallList(TerrainList);
 end;
// Pass 2 = Umrandungen
if pPassNr = 2 then
 begin
 glDisable(GL_TEXTURE_1D);
 glEnable(GL_CULL_FACE);
 glCullFace(GL_BACK);
 glLineWidth(4);
 glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
 // Vertexshader
 glBindProgramARB(GL_VERTEX_PROGRAM_ARB, VShader[1]);
 glEnable(GL_VERTEX_PROGRAM_ARB);
 // Terrain
 glCallList(TerrainList);
 glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
 end;
end;

// =============================================================================
//  TGLForm.ApplicationEventsIdle
// =============================================================================
procedure TGLForm.ApplicationEventsIdle(Sender: TObject; var Done: Boolean);
begin
glMatrixMode(GL_PROJECTION);
glLoadIdentity;
gluPerspective(90, ClientWidth/ClientHeight, 0.1, 512);
glViewport(0, 0, ClientWidth, ClientHeight);
glMatrixMode(GL_MODELVIEW);
glLoadIdentity;
glClear(GL_COLOR_BUFFER_BIT or GL_DEPTH_BUFFER_BIT);

glRotatef(Direction[0], 1,0,0);
glRotatef(Direction[1], 0,1,0);

// Skydome
glCallList(SkyDomeList);

glTranslatef(-Position[0], -Position[1], -Position[2]);

RenderPass(1);
RenderPass(2);

SwapBuffers(DC);
Done := False;
inc(Frame);

Position[0] := Position[0]+0.01*Sin(DegToRad(Direction[1]))*Move;
Position[1] := Position[1]-0.01*Sin(DegToRad(Direction[0]))*Move;
Position[2] := Position[2]-0.01*Cos(DegToRad(Direction[1]))*Move;
end;

// =============================================================================
//  TGLForm.FormMouseMove
// =============================================================================
procedure TGLForm.FormMouseMove(Sender: TObject; Shift: TShiftState; X,Y: Integer);
var
 MPos : TPoint;
begin
MPos := Mouse.CursorPos;
Direction[0] := Direction[0]+(MPos.Y-OldMPos.Y)*0.25;
Direction[1] := Direction[1]+(MPos.X-OldMPos.X)*0.25;
OldMPos := MPos;
if MPos.X < Left+15 then
 SetCursorPos(Left+ClientWidth-17, MPos.Y);
if MPos.X > Left+ClientWidth-15 then
 SetCursorPos(Left+17, MPos.Y);
if MPos.Y < Top+35 then
 SetCursorPos(MPos.X, Top+ClientHeight-17);
if MPos.Y > Top+ClientHeight-15 then
 SetCursorPos(MPos.X, Top+37);
OldMPos := Mouse.CursorPos;
end;

// =============================================================================
//  TGLForm.FPSTimerTimer
// =============================================================================
procedure TGLForm.FPSTimerTimer(Sender: TObject);
begin
Caption := 'CelShading-Demo 2003 by Sascha Willems ['+IntToStr(Frame)+'fps]';
Frame   := 0;
end;

// =============================================================================
//  TGLForm.FormMouseDown
// =============================================================================
procedure TGLForm.FormMouseDown(Sender: TObject; Button: TMouseButton;Shift: TShiftState; X, Y: Integer);
begin
case Button of
 mbLeft  : Move := 1;
 mbRight : Move := -1;
end;
end;

// =============================================================================
//  TGLForm.FormMouseUp
// =============================================================================
procedure TGLForm.FormMouseUp(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
Move := 0;
end;

procedure TGLForm.FormKeyPress(Sender: TObject; var Key: Char);
begin
if Key = #27 then
 Close;
end;

end.
