//24.09.2003 by Noeska Software (uses pivot points to draw meshes when needed, removed matrix)
//21.09.2003 by Noeska Software (more then 1 material per submesh, smoothed normals, dummy submeshes)
//18.09.2003 by Noeska Software (rewrite to 1 chunk parser) now it also reads non-standard 3ds files.
//16.09.2003 by Noeska Software
//08.09.2002 by Noeska Software
//01.05.2002 by Jan Michalowsky
//Thx a lot  Prometheus,Poetikus
//Version 1.02

unit gl3ds;

interface
uses Classes, SysUtils, dglOpenGL, GLMath;

type

  //types for reading of and working with 3ds files.
  TMap         = packed record
  tu,tv        :Single;
  end;

  T3DSMesh     = packed record
  name         :string[20]; //meshname
  version      :longword;
  matname      :array of string; // array of material names. (mostly 1)
  numvertex    :longword; // number of vertexes
  numindices   :longword; // number of faces
  pivot        :T3DPoint; //where to draw submesh (also for anim?)
  vertex       :array of T3dpoint;  //vertexes
  vnormal      :array of T3dpoint;  //vertex normals for smoother meshes
  mapping      :array of TMap; // uvmapping
  matid        :array of word; // material for a face
  indices      :array of word;  //faces
  end;

  T3DSMaterial  = packed record
  hastexturemap,
  hasmaterial,
  IsAmbient,
  IsDiffuse,
  IsSpecular    :boolean;
  name,
  filename      :string[255];
  shininess,
  us,vs,
  uoff,voff,rot :single;
  AmbR,AmbG,AmbB,
  DifR,DifG,DifB,
  SpcR,SpcG,SpcB:single;
  texid: integer;
  end;

  TAll3DSMesh  = packed record
  Name          :string[255];
  Mesh          :array of T3DSMesh;
  Material      :array of T3DSMaterial;
  minimum       :T3DPoint;
  maximum       :T3DPoint;
  Version       :longword;
  SubVersion    :longword;
  NumMeshes     :longword;
  NumMaterials  :longword;
  end;

function ReadMesh3DS(filename:string):TAll3DSMesh;
function ReadMesh3DSStream(_3DFile: Tstream):TAll3DSMesh;

procedure render3ds(mesh: TAll3DSMesh);
procedure rendermesh(lookin: TAll3DSMesh;mesh: T3DSMesh);

function GetMeshByName(lookin: TAll3DSMesh;s:string):T3DSMesh;
function GetMaterialByName(lookin: TAll3DSMesh;s:string): T3DSMaterial;
function GetMaterialById(lookin: TAll3DSMesh;id:word): T3DSMaterial;

implementation

const
     OVersion    = 3; //expects 3ds version 3 files

//Chunk ID's

     MAIN3DS     = $4D4D;
     VERS3DS     = $0002;
     EDIT3DS     = $3D3D;
     KEYF3DS     = $B000;

     EDIT_OBJECT = $4000;

     OBJ_HIDDEN  = $4010;
     OBJ_VERINFO = $3D3E;
     OBJ_TRIMESH = $4100;

     TRI_VERTEXL = $4110;
     TRI_FACEL1  = $4120;
     TRI_MAP     = $4130;
     TRI_MAPPING = $4140;
     TRI_MATRIX  = $4160;

     MAT_MATRIAL = $AFFF;
     MAT_MATNAME = $A000;
     MAT_AMBIENT = $A010;
     MAT_DIFFUSE = $A020;
     MAT_SPECULAR= $A030;
     MAT_TEXTURE = $A200;
     MAT_BUMPMAP = $A230;
     MAT_MAPFILE = $A300;
     MAT_VSCALE  = $A354;
     MAT_USCALE  = $A356;
     MAT_VOFF    = $A35A;
     MAT_UOFF    = $A358;
     MAT_TEXROT  = $A35C;
     MAT_COLOR   = $0010;
     MAT_COLOR24 = $0011;

     KEYF_OBJDES        = $B002;
     KEYF_OBJHIERARCH   = $B010;
     KEYF_OBJPIVOT      = $B013;

type
  T3DSColor=packed record
   r,g,b:single;
  end;

  rgb=packed record
   r,g,b:byte
  end;

  TChunkHdr    = packed record
   chunkID      :Word;
   chunkLE      :longword;
  end;

var meshindex:word  = 0;
    acount   :longword = 0;
    vcount   :longword = 0;
    mcount   :longword = 0;
    HChunk   :TAll3DSMesh;

{------------------------------------------------------------------}
{  Calculate normals for smoothing 3ds file                                                 }
{------------------------------------------------------------------}
procedure calcvnormals;
var
 f, m, found,count: integer;
 vertexn:t3dpoint;
begin

for m:=0 to hchunk.NumMeshes-1 do
 begin
 if hchunk.Mesh[m].numindices > 0 then
 begin
 SetLength(hchunk.Mesh[m].vnormal,hchunk.Mesh[m].numvertex);
 f:=0;
 while f < hchunk.Mesh[m].numindices-1 do // go through all vertexes and
  begin
    vertexn:=CalcNormalVector(hchunk.Mesh[m].vertex[hchunk.Mesh[m].indices[f]],hchunk.Mesh[m].vertex[hchunk.Mesh[m].indices[f+1]],hchunk.Mesh[m].vertex[hchunk.Mesh[m].indices[f+2]]);
    hchunk.Mesh[m].vnormal[hchunk.Mesh[m].indices[f]]:=vertexn;
    hchunk.Mesh[m].vnormal[hchunk.Mesh[m].indices[f+1]]:=vertexn;
    hchunk.Mesh[m].vnormal[hchunk.Mesh[m].indices[f+2]]:=vertexn;
    f:=f+3;
  end;
 end;
 end;
end;

procedure movetopivot;
var
 f, m, found,count: integer;
 pivot:t3dpoint;
begin

for m:=0 to hchunk.NumMeshes-1 do
 begin
 if hchunk.Mesh[m].numvertex > 0 then
 begin
 f:=0;
 while f < hchunk.Mesh[m].numvertex do // go through all vertexes and
  begin
    pivot.x:=hchunk.Mesh[m].pivot.x;
    pivot.y:=hchunk.Mesh[m].pivot.y;
    pivot.z:=hchunk.Mesh[m].pivot.z;

    hchunk.Mesh[m].vertex[f].x:=hchunk.Mesh[m].vertex[f].x-pivot.x;
    hchunk.Mesh[m].vertex[f].y:=hchunk.Mesh[m].vertex[f].y-pivot.y;
    hchunk.Mesh[m].vertex[f].z:=hchunk.Mesh[m].vertex[f].z-pivot.z;

    f:=f+1;
  end;
 end;
 end;
end;

procedure Calculatesize;
var m, f: integer;
    x,y,z: single;
begin
Hchunk.minimum.x:=0;
Hchunk.minimum.y:=0;
Hchunk.minimum.z:=0;
Hchunk.maximum.x:=0;
Hchunk.maximum.y:=0;
Hchunk.maximum.z:=0;
for m:=0 to Hchunk.NumMeshes-1 do
    for f:=0 to Hchunk.Mesh[m].NumIndices-1 do begin
        x:=Hchunk.Mesh[m].vertex[Hchunk.Mesh[m].indices[f]].x;
        y:=Hchunk.Mesh[m].vertex[Hchunk.Mesh[m].indices[f]].y;
        z:=Hchunk.Mesh[m].vertex[Hchunk.Mesh[m].indices[f]].z;
        if x<Hchunk.minimum.x then Hchunk.minimum.x:=x;
        if y<Hchunk.minimum.y then Hchunk.minimum.y:=y;
        if z<Hchunk.minimum.z then Hchunk.minimum.z:=z;
        if x>Hchunk.maximum.x then Hchunk.maximum.x:=x;
        if y>Hchunk.maximum.y then Hchunk.maximum.y:=y;
        if z>Hchunk.maximum.z then Hchunk.maximum.z:=z;
    end;
end;

function SwitchAxesP (const point:T3DPoint):T3DPoint;
begin
  with Result do begin
    x:=point.x; y:=point.z; z:=-point.y;
  end;
end;

function GetMeshByName(lookin: TAll3DSMesh;s:string):T3DSMesh;
var i:word;
begin
for i:=0 to High(lookin.Mesh) do
    if uppercase(lookin.Mesh[i].name)=uppercase(s) then
       begin
       result:=lookin.Mesh[i];
       break;
       end;
end;

function GetMaterialByName(lookin: TAll3DSMesh;s:string):T3DSMaterial;
var i:word;
begin

if lookin.nummaterials > 0 then
begin
for i:=0 to High(lookin.Material) do
    begin
    if uppercase(lookin.Material[i].name)=uppercase(s) then
       begin
       result:=lookin.Material[i];
       break;
       end;
    end;
end;
end;

function GetMaterialByID(lookin: TAll3DSMesh;id:word):T3DSMaterial;
var i:word;
begin
    result:=lookin.Material[id];
end;

function GetMaterialIDByName(s:string):word;
var i:word;
begin

if hchunk.nummaterials > 0 then
begin
for i:=0 to High(hchunk.Material) do
    begin
    if uppercase(hchunk.Material[i].name)=uppercase(s) then
       begin
       result:=i;
       break;
       end;
    end;
end;
end;


function ReadMesh3DS(filename:string):TAll3DSMesh;
var   buffer :TChunkHdr;
    _3DFile  :TFilestream;
begin
_3DFile:=TFilestream.Create(Filename,$0000);
result:=ReadMesh3DSStream(_3DFile);
if result.Version <> 3 then raise Exception.Create('Wrong 3DS Version');
_3DFile.Free;
end;

function ReadMesh3DSStream(_3DFile: Tstream):TAll3DSMesh;
var
    buffer :TChunkHdr;
    chrbuf:byte;
    color:t3dscolor;
    rgbcolor: rgb;
    count,i, dummy: word;
    tv     :T3dPoint;
    tm     :TMap;
    matarr :byte;
    matcount,matcountloop,mati, keyfcount: word;

begin

//init hchunk
HChunk.Name:='';
HChunk.Mesh:=nil;
HChunk.Material:=nil;
HChunk.NumMeshes:=0; // set number of meshes to zero
HChunk.NumMaterials:=0; // set number of materials to zero
acount:=0;
vcount:=0;
mcount:=1;
keyfcount:=0;

//start reading chunks
//now all is read in once... This helps for some none standard 3ds files like those generated by cinema4d.
while _3DFile.Position<_3DFile.Size do
      begin
      _3DFile.Read(buffer,sizeof(TChunkHdr));
         case buffer.chunkID of
         MAIN3DS:begin
                 //No nothing seems to be needed here...
                 end;
         VERS3DS:begin
                   _3DFile.Read(HChunk.Version,4);
                   if not (HChunk.Version=3) then
                     begin
                        _3DFile.Seek(0,3);
                        //???????
                     end;
                 end;
         EDIT3DS:begin
                 //No nothing seems to be needed here...
                 end;

         //read matrial data

         MAT_MATRIAL:begin
                     //inc(mcount);
                     //setlength(HChunk.Material,mcount);
                     //moved to MAT_MATNAME (thanks to non-standard 3ds files)
                     end;
        MAT_MATNAME:begin
                     inc(mcount);
                     setlength(HChunk.Material,mcount);
                     HChunk.NumMaterials:=HChunk.NumMaterials+1;

                     HChunk.Material[mcount-1].name:='';
                     chrbuf:=1;
                     while chrbuf<>0 do
                          begin
                          _3DFile.Read(chrbuf,1);
                          HChunk.Material[mcount-1].name:=HChunk.Material[mcount-1].name+chr(chrbuf);
                          end;
                     delete(HChunk.Material[mcount-1].name,length(HChunk.Material[mcount-1].name),1);
                     end;
         MAT_AMBIENT:begin
                     with HChunk.Material[mcount-1] do
                          begin
                          IsAmbient:=true;
                          IsDiffuse:=false;
                          IsSpecular:=false;
                          hasmaterial:=true;
                          end;
                     end;
         MAT_DIFFUSE:begin
                     with HChunk.Material[mcount-1] do
                          begin
                          IsAmbient:=false;
                          IsDiffuse:=true;
                          IsSpecular:=false;
                          hasmaterial:=true;
                          end;
                     end;
         MAT_SPECULAR:begin
                     with HChunk.Material[mcount-1] do
                          begin
                          IsAmbient:=false;
                          IsDiffuse:=false;
                          IsSpecular:=true;
                          hasmaterial:=true;
                          end;
                     end;
         MAT_TEXTURE:begin
                     with HChunk.Material[mcount-1] do
                          begin
                          hastexturemap:=true;
                          us:=0;
                          vs:=0;
                          uoff:=0;
                          voff:=0;
                          end;
                     end;
         MAT_MAPFILE:begin
                     chrbuf:=1;
                     HChunk.Material[mcount-1].filename:='';
                     while chrbuf<>0 do
                          begin
                          _3DFile.Read(chrbuf,1);

                          HChunk.Material[mcount-1].filename:=HChunk.Material[mcount-1].filename+chr(chrbuf);
                          end;
                     delete(HChunk.Material[mcount-1].filename,length(HChunk.Material[mcount-1].filename),1);
                     end;
         MAT_VSCALE:_3DFile.Read(HChunk.Material[mcount-1].vs,4);
         MAT_USCALE:_3DFile.Read(HChunk.Material[mcount-1].us,4);
         MAT_VOFF  :_3DFile.Read(HChunk.Material[mcount-1].voff,4);
         MAT_UOFF  :_3DFile.Read(HChunk.Material[mcount-1].uoff,4);
         MAT_TEXROT:_3DFile.Read(HChunk.Material[mcount-1].rot,4);
         MAT_COLOR24 :begin
                    _3DFile.Read(rgbcolor,sizeof(rgbcolor));
                    if HChunk.Material[mcount-1].IsDiffuse then
                       begin
                       with HChunk.Material[mcount-1] do
                            begin
                            difr:=rgbcolor.r / 256;
                            difg:=rgbcolor.g / 256;
                            difb:=rgbcolor.b / 256;
                            end;
                       end;
                    if HChunk.Material[mcount-1].IsAmbient then
                       begin
                       with HChunk.Material[mcount-1] do
                            begin
                            Ambr:=rgbcolor.r / 256;
                            Ambg:=rgbcolor.g / 256;
                            Ambb:=rgbcolor.b / 256;
                            end;
                       end;
                    if HChunk.Material[mcount-1].IsSpecular then
                       begin
                       with HChunk.Material[mcount-1] do
                            begin
                            Spcr:=rgbcolor.r / 256;
                            Spcg:=rgbcolor.g / 256;
                            Spcb:=rgbcolor.b / 256;
                            end;
                       end;
                    end;
         MAT_COLOR :begin
                    _3DFile.Read(color,sizeof(t3dscolor));
                    if HChunk.Material[mcount-1].IsDiffuse then
                       begin
                       with HChunk.Material[mcount-1] do
                            begin
                            difr:=color.r;
                            difg:=color.g;
                            difb:=color.b;
                            end;
                       end;
                    if HChunk.Material[mcount-1].IsAmbient then
                       begin
                       with HChunk.Material[mcount-1] do
                            begin
                            Ambr:=color.r;
                            Ambg:=color.g;
                            Ambb:=color.b;
                            end;
                       end;
                    if HChunk.Material[mcount-1].IsSpecular then
                       begin
                       with HChunk.Material[mcount-1] do
                            begin
                            Spcr:=color.r;
                            Spcg:=color.g;
                            Spcb:=color.b;
                            end;
                       end;

                    end;

        //read submeshes (objects) ...

         EDIT_OBJECT{, OBJ_HIDDEN}:begin
                     inc(acount);
                     setlength(HChunk.Mesh,acount);
                     chrbuf:=1;
                     HChunk.Mesh[acount-1].name:='';
                     while chrbuf<>0 do
                      begin
                       _3DFile.Read(chrbuf,1);
                       HChunk.Mesh[acount-1].name:=HChunk.Mesh[acount-1].name+chr(chrbuf);
                      end;
                     delete(HChunk.Mesh[acount-1].name,length(HChunk.Mesh[acount-1].name),1);
                     end;
         OBJ_TRIMESH:Begin
                        matarr:=0; //reset matarr to 0 for every submesh
                        HChunk.NumMeshes:=HChunk.NumMeshes+1;
                     End;

         //ReadVertices...

         TRI_VERTEXL:begin
                     _3DFile.Read(count,2);

                     if count > 0 then
                     begin
                       HChunk.Mesh[acount-1].numvertex:=count;
                       SetLength(HChunk.Mesh[acount-1].vertex,count);
                       for i:=0 to count-1 do
                         begin
                         _3DFile.Read(tv,12);
                         HChunk.Mesh[acount-1].vertex[i]:=tv;
                         end;
                       end
                       else
                       begin
                       HChunk.Mesh[acount-1].numvertex:=0;
                       end;
                     end;
         TRI_MAPPING:begin
                     _3DFile.Read(count,2);
                     SetLength(HChunk.Mesh[acount-1].mapping,count);
                     for i:=0 to count-1 do
                         begin
                         _3DFile.Read(tm,8);
                         HChunk.Mesh[acount-1].mapping[i]:=tm;
                         end;
                     end;
         TRI_MAP    :begin
                     chrbuf:=1;
                     inc(matarr);
                     setlength(HChunk.Mesh[acount-1].matname,matarr);
                     HChunk.Mesh[acount-1].matname[matarr-1]:='';
                     while chrbuf<>0 do
                          begin
                          _3DFile.Read(chrbuf,1);
                          HChunk.Mesh[acount-1].matname[matarr-1]:=HChunk.Mesh[acount-1].matname[matarr-1]+chr(chrbuf);
                          end;

                     delete(HChunk.Mesh[acount-1].matname[matarr-1],length(HChunk.Mesh[acount-1].matname[matarr-1]),1);

                     //look up and set matid for vertices with this material
                     _3DFile.Read(matcount,2);
                     SetLength(HChunk.Mesh[acount-1].matid,HChunk.Mesh[acount-1].numindices div 3);
                     for matcountloop:=0 to matcount-1 do
                      begin
                        _3DFile.Read(mati,2);
                        HChunk.Mesh[acount-1].matid[mati]:=GetMaterialIDbyName(HChunk.Mesh[acount-1].matname[matarr-1]);
                      end;

                     end;
         TRI_FACEL1 :begin
                     _3DFile.Read(count,2);
                     SetLength(HChunk.Mesh[acount-1].indices,count*3);
                     HChunk.Mesh[acount-1].numindices:=count*3;
                     for i:=0 to count-1 do
                             begin
                             _3DFile.Read(HChunk.Mesh[acount-1].indices[i*3+2],2);
                             _3DFile.Read(HChunk.Mesh[acount-1].indices[i*3+1],2);
                             _3DFile.Read(HChunk.Mesh[acount-1].indices[i*3],2);
                             _3DFile.Read(dummy,2);
                             end;
                     end;

         // read in keyframe data if available (for now only for pivot)
         KEYF3DS     :begin
                        //No nothing seems to be needed here...
                      end;
         KEYF_OBJDES :begin
                        //if all is ok then for every submesh a keyf_objdes exists
                        inc(keyfcount);
                      end;
         KEYF_OBJPIVOT :begin
                                //read in pivot point
                                _3DFile.Read(tv,12);
                                HChunk.Mesh[keyfcount-1].pivot:=tv;
                        end;
         else
         _3DFile.Seek(buffer.chunkLE-sizeof(TChunkHdr),soFromCurrent);
         end;
      end;

//calculate some things afterwards...
movetopivot; //if pivots are used in 3ds move submeshes
calcvnormals; //calculate vertex normals for a smooth looking 3ds
Calculatesize; //calculate min and max size

result:=HChunk;

end;

{------------------------------------------------------------------}
{  Render 3ds file                                                 }
{------------------------------------------------------------------}
procedure render3ds(mesh: TAll3DSMesh);
var
 m: integer{dword};
 texture: gluint;

begin

 for m:=0 to mesh.NumMeshes-1 do
 begin
   rendermesh(mesh,mesh.Mesh[m]);
 end;

end;

procedure rendermesh(lookin: TAll3dsMesh;mesh: T3DSMesh);
var
 f: integer;
 mat: T3DSMaterial;
 mc: integer;
 ambient,diffuse,specular, emissive: TGLCOLOR;
 power:single;
 vertex, vertex2, vertex3, vnormal, vnormal2, vnormal3:T3DPOINT;

begin

  if Mesh.numindices > 0 then
  begin

   f:=0;
   while f< mesh.numindices-1 do
    begin

     //begin setting material

     mat:=GetMaterialByID(lookin,mesh.matid[f div 3]);
     //gldisable(GL_TEXTURE_2D); //just to make sure...
     diffuse.red:=mat.DifR;
     diffuse.green:=mat.DifG;
     diffuse.blue:=mat.DifB;
     diffuse.alpha:=1.0;

     //if no ambient color data then also set diffuse for ambient
     if mat.IsAmbient then
      begin
       ambient.red:=mat.AmbR;
       ambient.green:=mat.AmbG;
       ambient.blue:=mat.AmbB;
       ambient.alpha:=1.0;
       glColor3f(ambient.red, ambient.green, ambient.blue);
      end
     else
      begin
       ambient.red:=mat.DifR;
       ambient.green:=mat.DifG;
       ambient.blue:=mat.DifB;
       ambient.alpha:=1.0;
       glColor3f(ambient.red, ambient.green, ambient.blue);
      end;

     specular.red:=mat.spcR;
     specular.green:=mat.spcG;
     specular.blue:=mat.spcB;
     specular.alpha:=1.0;

     with emissive do begin red:=0.0;green:=0.0;blue:=0.0;alpha:=1.0;end;
     power:=1.0;

     {glMaterialfv(GL_FRONT,gl_ambient,@ambient);
     glMaterialfv(GL_FRONT,gl_diffuse,@diffuse);
     glMaterialfv(GL_FRONT,gl_specular,@specular);
     glMaterialfv(GL_FRONT,gl_shininess,@power);
     glMaterialfv(GL_FRONT,gl_emission,@emissive);}

     //end setting material

     vertex:=SwitchAxesP(mesh.vertex[mesh.indices[f]]);
     vertex2:=SwitchAxesP(mesh.vertex[mesh.indices[f+1]]);
     vertex3:=SwitchAxesP(mesh.vertex[mesh.indices[f+2]]);

     vnormal:=mesh.vnormal[mesh.indices[f]];
     vnormal2:=mesh.vnormal[mesh.indices[f+1]];
     vnormal3:=mesh.vnormal[mesh.indices[f+2]];

     glbegin(GL_TRIANGLE_FAN);
      glNormal3f(vnormal.x,vnormal.y,vnormal.z);
      glVertex3f(vertex.x,vertex.y,vertex.z);
      glNormal3f(vnormal2.x,vnormal2.y,vnormal2.z);
      glVertex3f(vertex2.x,vertex2.y,vertex2.z);
      glNormal3f(vnormal3.x,vnormal3.y,vnormal3.z);
      glVertex3f(vertex3.x,vertex3.y,vertex3.z);
      f:=f+3;
     glend;

    end;
   end;

end;

end.
