{ ****************************************************************************** }
{                                                                                }
{  Motionblur demo -  2003 by Sascha Willems                                    }
{   visit http://www.delphigl.de for more                                        }
{   visit http://www.delphigl.com the biggest german Delphi+OpenGL community     }
{                                                                                }
{  This demo uses the accumulationbuffer to simulate a motion blur.              }
{  Note that the accum-buffer isn't supported in hardware on most 3d-cards       }
{  except ATI's R300.So you won't get acceptable framerates on any other cards   }
{  tham a Radeon9500/9600/9700/9800                                              }
{                                                                                }
{ ****************************************************************************** }
{  Compiled with Delphi 7.0                                                      }
{ ****************************************************************************** }
unit MotionBlur1;

interface

uses
  Windows,
  Messages,
  SysUtils,
  Variants,
  Classes,
  Graphics,
  Controls,
  Forms,
  Dialogs,

  dglOpenGL,
  glo,
  Textures,
  AppEvnts,
  ExtCtrls;

type
  TForm1 = class(TForm)
    ApplicationEvents1: TApplicationEvents;
    RotTimer: TTimer;
    FPSTimer: TTimer;
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure ApplicationEvents1Idle(Sender: TObject; var Done: Boolean);
    procedure RotTimerTimer(Sender: TObject);
    procedure FPSTimerTimer(Sender: TObject);
    procedure FormKeyPress(Sender: TObject; var Key: Char);
  private
    { Private-Deklarationen }
  public
    { Public-Deklarationen }
   DC  : HDC;
   RC  : HGLRC;
   Rot : Single;
   procedure DrawSkyBox;
   procedure DrawTunnel(pAngle : Single);
  end;

const
 TexturePath = 'textures\';
 TextureList : array[0..0] of String = ('doom.jpg');

var
  Form1      : TForm1;
  Textures   : array of TGLUInt;
  Frames     : Integer;
  NumSamples : Integer;
  NumDegrees : Integer;
  UseAccum   : Boolean;

  Tunnels    : array[0..32, 0..32, 0..3] of TGLFloat;
  Angle      : TGLFloat;

implementation

{$R *.dfm}

procedure TForm1.FormCreate(Sender: TObject);
var
 i : Integer;
begin
ChDir(ExtractFilePath(Application.ExeName));
InitOpenGL;
DC := GetDC(Handle);
RC := CreateRenderingContext(DC, [opDoubleBuffered], 32, 32, 0, 32, 0, 0);
ActivateRenderingContext(DC, RC);
glEnable(GL_TEXTURE_2D);
glEnable(GL_DEPTH_TEST);
glDepthFunc(GL_LESS);
SetLength(Textures, Length(TextureList));
for i := 0 to High(TextureList) do
 LoadTexture(TexturePath+TextureList[i], Textures[i], False, GL_LINEAR, GL_LINEAR_MIPMAP_LINEAR, False);
NumSamples := 10;
NumDegrees := 2;
end;

procedure TForm1.FormDestroy(Sender: TObject);
begin
DeActivateRenderingContext;
wglDeleteContext(RC);
ReleaseDC(Handle, DC);
end;

procedure TForm1.DrawTunnel(pAngle : Single);
const
 TEXTURE_SPEED = 1/50;
var
 i,j     : Integer;
 c,j1,j2 : TGLFloat;
begin
// setup tunnel coordinates
for i := 0 to 12 do
 for j := 0 to 32 do
  begin
  Tunnels[I, J][0] :=(3 - J/12)*cos(2*pi/12*I) + 2*sin((Angle+2*j)/29) + cos((Angle+2*j)/13) - 2*sin(Angle/29) - cos(Angle/13);
  Tunnels[I, J][1] :=(3 - J/12)*sin(2*pi/12*I) + 2*cos((Angle+2*j)/33) + sin((Angle+2*j)/17) - 2*cos(Angle/33) - sin(Angle/17);
  Tunnels[I, J][2] :=-J;
  end;
// draw tunnel
for j := 0 to 30 do
 begin
 J1 :=J/32 + pAngle*TEXTURE_SPEED;        // precalculate texture v coords for speed
 J2 :=(J+1)/32 + pAngle*TEXTURE_SPEED;
 // near the end of the tunnel, fade the effect away
 if J > 24 then
  C :=1.0-(J-24)/10
 else
  C :=1.0;
 glColor3f(C, C, C);
 glBegin(GL_QUADS);
  for I :=0 to 11 do
   begin
   glTexCoord2f((I-3)/12, J1); glVertex3f(Tunnels[ I,   J ][0], Tunnels[ I,   J ][1], Tunnels[ I,   J ][2]);
   glTexCoord2f((I-2)/12, J1); glVertex3f(Tunnels[I+1,  J ][0], Tunnels[I+1,  J ][1], Tunnels[I+1,  J ][2]);
   glTexCoord2f((I-2)/12, J2); glVertex3f(Tunnels[I+1, J+1][0], Tunnels[I+1, J+1][1], Tunnels[I+1, J+1][2]);
   glTexCoord2f((I-3)/12, J2); glVertex3f(Tunnels[ I,  J+1][0], Tunnels[ I,  J+1][1], Tunnels[ I,  J+1][2]);
   end;
 glEnd;
 end;
end;

procedure TForm1.DrawSkyBox;
begin
glPushMatrix;
glTranslatef(-0.5, -0.5, -0.5);
 // Back
 glBindTexture(GL_TEXTURE_2D, Textures[1]);
 glBegin(GL_QUADS);
  glTexCoord2f(0, 0); glVertex3f(0, 0, 0);
  glTexCoord2f(0, 1); glVertex3f(0, 1, 0);
  glTexCoord2f(1, 1); glVertex3f(1, 1, 0);
  glTexCoord2f(1, 0); glVertex3f(1, 0, 0);
 glEnd;
 // Front
 glBindTexture(GL_TEXTURE_2D, Textures[0]);
 glBegin(GL_QUADS);
  glTexCoord2f(1, 0); glVertex3f(0,	0, 1);
  glTexCoord2f(1, 1); glVertex3f(0, 1, 1);
  glTexCoord2f(0, 1); glVertex3f(1, 1, 1);
  glTexCoord2f(0, 0); glVertex3f(1, 0, 1);
 glEnd;
 // Bottom
 glBindTexture(GL_TEXTURE_2D, Textures[5]);
 glBegin(GL_QUADS);
  glTexCoord2f(0, 0); glVertex3f(0,	0, 0);
  glTexCoord2f(0, 1); glVertex3f(0,	0, 1);
  glTexCoord2f(1, 1); glVertex3f(1, 0, 1);
  glTexCoord2f(1, 0); glVertex3f(1, 0, 0);
 glEnd;
 // Top
 glBindTexture(GL_TEXTURE_2D, Textures[4]);
 glBegin(GL_QUADS);
  glTexCoord2f(0, 0); glVertex3f(0, 1, 0);
  glTexCoord2f(0, 1); glVertex3f(0, 1, 1);
  glTexCoord2f(1, 1); glVertex3f(1, 1, 1);
  glTexCoord2f(1, 0); glVertex3f(1, 1, 0);
 glEnd;
 // Left
 glBindTexture(GL_TEXTURE_2D, Textures[2]);
 glBegin(GL_QUADS);
  glTexCoord2f(1, 0); glVertex3f(0, 0, 0);
  glTexCoord2f(0, 0); glVertex3f(0, 0, 1);
  glTexCoord2f(0, 1); glVertex3f(0, 1, 1);
  glTexCoord2f(1, 1); glVertex3f(0, 1, 0);
 glEnd;
 // Right
 glBindTexture(GL_TEXTURE_2D, Textures[3]);
 glBegin(GL_QUADS);
  glTexCoord2f(0, 0); glVertex3f(1, 0, 0);
  glTexCoord2f(1, 0); glVertex3f(1, 0, 1);
  glTexCoord2f(1, 1); glVertex3f(1, 1, 1);
  glTexCoord2f(0, 1); glVertex3f(1, 1, 0);
 glEnd;
glPopMatrix;
end;

procedure DrawScene(pRot,pAngle,pOffset : Single);
begin
glLoadIdentity;

glPushMatrix;
 glTranslatef(0,0,pOffset);
 glEnable(GL_TEXTURE_2D);
 glBindTexture(GL_TEXTURE_2D, Textures[0]);
 glPushMatrix;
  Form1.DrawTunnel(pAngle);
 glPopMatrix;
glPopMatrix;

glDisable(GL_TEXTURE_2D);
end;

procedure TForm1.ApplicationEvents1Idle(Sender: TObject; var Done: Boolean);
var
 i : Integer;
begin
Angle := Angle + 0.5;
glMatrixMode(GL_PROJECTION);
glLoadIdentity;
gluPerspective(60, ClientWidth/ClientHeight, 0.1, 256);
glViewPort(0,0,ClientWidth,ClientHeight);
glMatrixMode(GL_MODELVIEW);
glClear(GL_ACCUM_BUFFER_BIT);
// Szene "lschen"
glClear(GL_DEPTH_BUFFER_BIT or GL_COLOR_BUFFER_BIT);
DrawScene(Rot, Angle, 0);
glAccum(GL_LOAD, 0.5);
// Motionblur
for i := 1 to NumSamples-1 do
 begin
 glClear(GL_DEPTH_BUFFER_BIT or GL_COLOR_BUFFER_BIT);
 DrawScene(Rot-i*NumDegrees, Angle-i*NumDegrees/4, i);
 glAccum(GL_ACCUM, 1/i/2);
 end;
// Akkumulationspuffer in den Backpuffer kopieren
glAccum(GL_RETURN, 1);
// Backpuffer nach vorne bringen
SwapBuffers(DC);
inc(Frames);
Done := False;
end;

procedure TForm1.RotTimerTimer(Sender: TObject);
begin
Rot := Rot+2;
if Rot > 360 then
 Rot := 0;
end;

procedure TForm1.FPSTimerTimer(Sender: TObject);
begin
Caption := 'MotionBlur-Demo - '+IntToStr(Frames)+'fps';
Frames  := 0;
end;

procedure TForm1.FormKeyPress(Sender: TObject; var Key: Char);
begin
case Key of
 'a' : if NumSamples > 2 then
        dec(NumSamples);
 's' : if NumSamples < 20 then
        inc(NumSamples);
 '+' : if NumDegrees < 20 then
        inc(NumDegrees);
 '-' : if NumDegrees > 1 then
        dec(NumDegrees);
end;
end;

end.
