unit ProjectedTex1;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms,
  Dialogs, glo, OpenGL12, AppEvnts, ExtCtrls, Textures;

type
  TForm1 = class(TForm)
   ApplicationEvents1: TApplicationEvents;
    RotateTimer: TTimer;
    procedure FormResize(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure ApplicationEvents1Idle(Sender: TObject; var Done: Boolean);
    procedure RotateTimerTimer(Sender: TObject);
    procedure FormKeyPress(Sender: TObject; var Key: Char);
  private
    { Private-Deklarationen }
  public
   DC : HDC;
   RC : HGLRC;
   procedure Display;
   procedure DrawScene(pUseTextures : Boolean);
   procedure Init;
  end;
  TVector4f = array[0..3] of Single;

var
  Form1: TForm1;

const
 LV             : array[0..3] of TGLFloat = (0, 10, 0, 1);
 ProjectorPersp : Single = 45;

var
 Projector      : TGLUInt;
 Wall           : TGLUInt;
 Floor          : TGLUInt;
 Metal          : TGLUInt;
 WireFrame      : Boolean;
 ProjectorAngle : Single;

implementation

{$R *.dfm}

function Vector4f(px, py, pz, pw : Single) : TVector4f;
begin
Result[0] := px;
Result[1] := py;
Result[2] := pz;
Result[3] := pw;
end;

procedure TForm1.FormCreate(Sender: TObject);
var
 DummyPal : HPalette;
begin
if MessageDlg('Switch to fullscreen?', mtConfirmation, [mbYes, mbNo], 0) = mrYes then
 begin
 BorderStyle := bsNone;
 WindowState := wsMaximized;
 end;
DC := GetDC(Handle);
RC := CreateRenderingContext(DC, [opDoubleBuffered], 32, 24, 0, 0, 0, 0, DummyPal);
wglMakeCurrent(DC, RC);
Init;
Resize;
end;

procedure TForm1.FormResize(Sender: TObject);
begin
if (DC=0) or (RC=0) then
 exit;
glViewport(0, 0, ClientWidth, ClientHeight);
glMatrixMode(GL_PROJECTION);
glLoadIdentity;
gluPerspective(60, ClientWidth/ClientHeight, 1, 1000);
glMatrixMode(GL_MODELVIEW);
end;

procedure TForm1.ApplicationEvents1Idle(Sender: TObject; var Done: Boolean);
begin
if (DC=0) or (RC=0) then
 exit;
Display;
Done := False;
end;

procedure TForm1.Display;
var
 P : TVector4f;
begin
if WireFrame then
 glPolygonMode(GL_FRONT_AND_BACK, GL_LINE)
else
 glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);

glClear(GL_COLOR_BUFFER_BIT or GL_DEPTH_BUFFER_BIT);
glLoadIdentity;
// Betrachterposition und -blickwinkel setzen
glRotatef(20, 1, 0, 0);
glRotatef(-20, 0, 1, 0);
glTranslatef(0, 0, -50);
// Automatische Texturkoordinatengeneration deaktivieren
glDisable(GL_TEXTURE_GEN_S);
glDisable(GL_TEXTURE_GEN_T);
glDisable(GL_TEXTURE_GEN_R);
glDisable(GL_TEXTURE_GEN_Q);
// Erster Durchgang : Normale Szene zeichnen
// Texturenmatrix sichern und zurcksetzen
glMatrixMode(GL_TEXTURE);
 glPushMatrix;
 glLoadIdentity;
 // In die Modelansicht wechseln und Szene mit aktivierter Beleuchtung und Texturen zeichnen
 glMatrixMode(GL_MODELVIEW);
 glEnable(GL_LIGHTING);
 DrawScene(True);
 glEnable(GL_TEXTURE_2D);
 // Texturenmatrix (in der die Projektorperspektive gesichert wird) wieder zurckholen
 glMatrixMode(GL_TEXTURE);
 glPopMatrix;
glMatrixMode(GL_MODELVIEW);
glDisable(GL_LIGHTING);
// Texgen-Flchen setzen (werden mit der invertierten Modelansichtsmatrix multiplizert)
P := Vector4f(1,0,0,0); glTexGenfv(GL_S, GL_EYE_PLANE, @P);
P := Vector4f(0,1,0,0); glTexGenfv(GL_T, GL_EYE_PLANE, @P);
P := Vector4f(0,0,1,0); glTexGenfv(GL_R, GL_EYE_PLANE, @P);
P := Vector4f(0,0,0,1); glTexGenfv(GL_Q, GL_EYE_PLANE, @P);
// Automatische Texturkoordinatengeneration aktivieren
glEnable(GL_TEXTURE_GEN_S);
glEnable(GL_TEXTURE_GEN_T);
glEnable(GL_TEXTURE_GEN_R);
glEnable(GL_TEXTURE_GEN_Q);
// Bleding aktivieren damit der Projektor nicht die komplette Szene berzeichnet
glEnable(GL_BLEND);
glBlendFunc(GL_DST_COLOR, GL_ONE);
// In die Texturenmatrix wechseln und den Projektor ausrichten
glMatrixMode(GL_TEXTURE);
 glPopMatrix;
 glPushMatrix;
 glBindTexture(GL_TEXTURE_2D, Projector);
 // Position und Ausrichtung des Projektors festlegen
 glRotatef(-30+cos(ProjectorAngle*PI/180)*30, 0, 1, 0);
 glRotatef( 25+sin(ProjectorAngle*PI/180)*5, 1, 0, 0);
 glTranslatef(0, 0, -20);
// Szene mit projezierter Textur zeichnen
glMatrixMode(GL_MODELVIEW);
DrawScene(False);
glDisable(GL_BLEND);
SwapBuffers(DC);
end;

procedure TForm1.DrawScene(pUseTextures : Boolean);
var
 i : Integer;
begin
if pUseTextures then
 glBindTexture(GL_TEXTURE_2D, Metal);
// Gitterstbe
glPushMatrix;
 glTranslatef(-46, 15, -0);
 glRotatef(90, 1, 0, 0);
 for i := 0 to 11 do
  begin
  glTranslatef(6, 0, 0);
  gloSolidCylinder(1, 1, 55, 16, 1);
  end;
glPopMatrix;
glEnable(GL_TEXTURE_2D);
// Boden
if pUseTextures then
 glBindTexture(GL_TEXTURE_2D, Floor);
glBegin(GL_QUADS);
 glNormal3f(0, 1, 0);
 glTexCoord2f(0,0); glVertex3f(-40, -40, -40);
 glTexCoord2f(3,0); glVertex3f( 40, -40, -40);
 glTexCoord2f(3,3); glVertex3f( 40, -40,  40);
 glTexCoord2f(0,3); glVertex3f(-40, -40,  40);
glEnd;
// Wnde
if pUseTextures then
 glBindTexture(GL_TEXTURE_2D, Wall);
glBegin(GL_QUADS);
 glNormal3f(1, 0, 0);
 glTexCoord2f(0,0); glVertex3f(-40, -40, -40);
 glTexCoord2f(3,0); glVertex3f(-40, -40,  40);
 glTexCoord2f(3,3); glVertex3f(-40,  40,  40);
 glTexCoord2f(0,3); glVertex3f(-40,  40, -40);
 glNormal3f(0, 0, 1);
 glTexCoord2f(0,0); glVertex3f(-40, -40, -40);
 glTexCoord2f(3,0); glVertex3f( 40, -40, -40);
 glTexCoord2f(3,3); glVertex3f( 40,  40, -40);
 glTexCoord2f(0,3); glVertex3f(-40,  40, -40);
glEnd;

end;

procedure TForm1.Init;
const
 Ambient  : array[0..3] of TGLFLoat = (0.2, 0.2, 0.2, 1);
 Diffuse  : array[0..3] of TGLFLoat = (0.8, 0.8, 0.8, 1);
 Specular : array[0..3] of TGLFLoat = (0.5, 0.6, 0.8, 1);
begin
glClearColor(0.3, 0.3, 0.3, 1);
glClearDepth(1);
glEnable(GL_DEPTH_TEST);
glEnable(GL_BLEND);
glEnable(GL_LIGHTING);
glEnable(GL_LIGHT0);
glDepthFunc(GL_LEQUAL);
glEnable(GL_TEXTURE_2D);
// Projektortextur laden und wichtige Parameter setzen, damit dieser korrekt dargestellt wird
LoadTexture('projector.tga', Projector, False);
glBindTexture(GL_TEXTURE_2D, Projector);
glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
// Bodentexture
LoadTexture('floor.jpg', Floor, False);
glBindTexture(GL_TEXTURE_2D, Floor);
glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
// Wandtextur
LoadTexture('wall.jpg', Wall, False);
glBindTexture(GL_TEXTURE_2D, Wall);
glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
// Stabtextur
LoadTexture('metal.jpg', Metal, False);
glBindTexture(GL_TEXTURE_2D, Metal);
glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
// Lichteigenschaften setzen
glLightfv(GL_LIGHT0, GL_AMBIENT, @Ambient);
glLightfv(GL_LIGHT0, GL_SPECULAR, @Specular);
glLightfv(GL_LIGHT0, GL_DIFFUSE, @Diffuse);
glLightfv(GL_LIGHT0, GL_POSITION, @Lv);
// Projektor initialisieren
glMatrixMode(GL_TEXTURE);
glLoadIdentity;
// Texturenmatrix skalieren und an 0..1 anpassen
glTranslatef(0.5, 0.5, 0);
glScalef(0.5, 0.5, 1);
// Projekorperspektive setzen
gluPerspective(ProjectorPersp, 1, 0.1, 1);
// Matrix sichern, damit diese spter wieder ohne obigen Initialisierungsblock wieder aufgerufen werden kann
glPushMatrix;
glMatrixMode(GL_MODELVIEW);
// OpenGLs automatische Texturkoordinatengeneration aktivieren
glEnable(GL_TEXTURE_GEN_S);
glEnable(GL_TEXTURE_GEN_T);
glEnable(GL_TEXTURE_GEN_R);
glEnable(GL_TEXTURE_GEN_Q);
// Hier wird OpenGL mitgeteilt, das die Weltkoordinaten in Texturkoordinaten umgewandelt werden
glTexGeni(GL_S, GL_TEXTURE_GEN_MODE, GL_EYE_LINEAR);
glTexGeni(GL_T, GL_TEXTURE_GEN_MODE, GL_EYE_LINEAR);
glTexGeni(GL_R, GL_TEXTURE_GEN_MODE, GL_EYE_LINEAR);
glTexGeni(GL_Q, GL_TEXTURE_GEN_MODE, GL_EYE_LINEAR);
end;


procedure TForm1.RotateTimerTimer(Sender: TObject);
begin
// Projektorrotation
ProjectorAngle := ProjectorAngle+0.2;
if ProjectorAngle > 360 then
 ProjectorAngle := 0;
end;

procedure TForm1.FormKeyPress(Sender: TObject; var Key: Char);
begin
case Key of
 'w' : WireFrame := not WireFrame;
 #27 : Close;
end;
end;

end.

