// =============================================================================
//  Shadowmapping Demo - 2003 by Sascha Willems
//  visit http://www.delphigl.de for more
//
//  Delphi-Port der shadow mapping demo fr c++ von www.codesampler.com
//
//
//  Mehr Infos zu dieser Technik gibts in folgenden Dokumenten :
//    http://developer.nvidia.com/docs/IO/1379/ATT/GDC2K_ogl_shadowmap.pdf
//    http://developer.nvidia.com/docs/IO/1252/ATT/ShadowMapping.pdf
//
//
//  Hinweis : Diese Demo nutzt nVidia-spezifische Erweiterungen, und kann auf
//            Karten anderer Hersteller deshalb Probleme machen!
// =============================================================================

unit ShadowMap1;

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, ShadowMap_global, AppEvnts, OpenGL12, glExtensions, ExtCtrls, glo;

type
  TForm1 = class(TForm)
    ApplicationEvents1: TApplicationEvents;
    FPSTimer: TTimer;
    RotateTimer: TTimer;
    procedure FormCreate(Sender: TObject);
    procedure ApplicationEvents1Idle(Sender: TObject; var Done: Boolean);
    procedure FormKeyDown(Sender: TObject; var Key: Word;
      Shift: TShiftState);
    procedure FormKeyPress(Sender: TObject; var Key: Char);
    procedure FPSTimerTimer(Sender: TObject);
    procedure RotateTimerTimer(Sender: TObject);
  private
    { Private-Deklarationen }
  public
    { Public-Deklarationen }
  end;

var
  Form1: TForm1;

implementation

{$R *.dfm}

procedure TForm1.FormCreate(Sender: TObject);
begin
ShadowMapDemo.InitOpenGL(Handle);
ShadowMapDemo.ObjConfig := 4;
ShadowMapDemo.ObjRotate := True;
ShadowMapDemo.WireFrame := False;
SpinX_L                 := 135;
end;

procedure TForm1.ApplicationEvents1Idle(Sender: TObject; var Done: Boolean);
begin
WindowSize.x := ClientWidth;
WindowSize.y := ClientHeight;
ShadowMapDemo.Render;
Done := False;
inc(Frames);
end;

procedure TForm1.FormKeyDown(Sender: TObject; var Key: Word;
  Shift: TShiftState);
begin
case Key of
 VK_LEFT     : SpinX_L := SpinX_L-1;
 VK_RIGHT    : SpinX_L := SpinX_L+1;
 VK_DOWN     : LightPosition[1] := LightPosition[1]-0.1;
 VK_UP       : LightPosition[1] := LightPosition[1]+0.1;
 VK_ADD      : LightPosition[0] := LightPosition[0]-0.1;
 VK_SUBTRACT : LightPosition[0] := LightPosition[0]+0.1;
end;
end;

procedure TForm1.FormKeyPress(Sender: TObject; var Key: Char);
begin
case Key of
 't' : begin
       ShadowSize := ShadowSize*2;
       if ShadowSize > 1024 then
        ShadowSize := 32;
       ShadowMapDemo.CreateDepthTexture;
       end;
 'd' : RenderDepthTex := not RenderDepthTex;
 'r' : ShadowMapDemo.ObjRotate := not ShadowMapDemo.ObjRotate;
 'o' : if ShadowMapDemo.ObjConfig < 4 then
        ShadowMapDemo.ObjConfig := ShadowMapDemo.ObjConfig+1;
 'O' : if ShadowMapDemo.ObjConfig > 0 then
        ShadowMapDemo.ObjConfig := ShadowMapDemo.ObjConfig-1;
 'w' : ShadowMapDemo.WireFrame := not ShadowMapDemo.WireFrame;
 'h' : ShowMessage('Shadowmap Demo - 2003 by Sascha Willems'+#13+'visit http://www.delphigl.de for more'+#13+#13+'Keys:'+#13+
                   't - Change shadowtexture resolution (current:'+IntToStr(ShadowSize)+')'+#13+
                   'd - Show depth texture'+#13+'r - Toggle rotating objects'+#13+'o - Next object configuration'+#13+
                   'O - Previous object configuration'+#13+'w - Toggle wireframe'+#13+'+/- - Move light along x-axis'+#13+
                   'Cursor up/down - Move light along y-axis'+#13+'Cursor left/right - Rotate scene');
end;
end;

procedure TForm1.FPSTimerTimer(Sender: TObject);
begin
Caption := 'Shadowmapping Demo 2003 by Sascha Willems ['+IntToStr(Frames)+'fps]';
Frames  := 0;
end;

procedure TForm1.RotateTimerTimer(Sender: TObject);
begin
with ShadowMapDemo do
 if ObjRotate then
  begin
  Spin := Spin+0.2;
  if Spin > 360 then
   Spin := 0;
  end;
end;

end.
