// =============================================================================
//  Shadowmapping Demo - 2003 by Sascha Willems
//  visit http://www.delphigl.de for more
//
//  Delphi-Port der shadow mapping demo fr c++ von www.codesampler.com
//
//
//  Mehr Infos zu dieser Technik gibts in folgenden Dokumenten :
//    http://developer.nvidia.com/docs/IO/1379/ATT/GDC2K_ogl_shadowmap.pdf
//    http://developer.nvidia.com/docs/IO/1252/ATT/ShadowMapping.pdf
//
//
//  Hinweis : Diese Demo nutzt nVidia-spezifische Erweiterungen, und kann auf
//            Karten anderer Hersteller deshalb Probleme machen!
// =============================================================================

unit ShadowMap_global;

interface

uses
 Windows,
 glExtensions,
 Dialogs,
 SysUtils,
 OpenGL12,

 glo;

type
 TShadowMapDemo = object
   DC        : HDC;
   RC        : HGLRC;
   ObjConfig : Byte;
   ObjRotate : Boolean;
   WireFrame : Boolean;
   Spin      : Single;
   procedure InitOpenGL(pHandle : THandle);
   procedure DeInitOpenGL;
   procedure CreateDepthTexture;
   procedure UpdateDepthTexture;
   procedure DisplayDepthTexture;
   procedure RenderScene(pDrawFloor : Boolean);
   procedure Render;
  end;

const
 ColorRed      : array[0..3] of Single = (0.8, 0.2, 0.1, 1);
 ColorGreen    : array[0..3] of Single = (0.1, 0.7, 0.2, 1);
 ColorGray     : array[0..3] of Single = (0.4, 0.4, 0.4, 1);
 SpinX_L       : Single =   0;
 SpinY_L       : Single = -10;
 SpinX_R       : Single =   0;
 SpinY_R       : Single =   0;
 ClipNear      = 0.1;
 ClipFar       = 512;

var
 LightLookAtMatrix : array[0..15] of Single;
 LightPosition     : array[0..3] of Single;
 ShadowMapDemo     : TShadowMapDemo;
 DepthTexture      : TGLUInt;
 RenderDepthTex    : Boolean;
 WindowSize        : TPoint;
 Frames            : Integer;
 ShadowSize        : Integer;

implementation

procedure TShadowMapDemo.InitOpenGL(pHandle : THandle);
const
 LightAmbient   : array[0..3] of Single = (0.25, 0.25, 0.25, 1);
 LightDiffuse   : array[0..3] of Single = (1, 1, 1, 1);
 LightLinearAtt : Single = 0;
var
 DummyPal : HPalette;
begin
LightPosition[0] := 2;
LightPosition[1] := 6.5;
LightPosition[2] := 0;
LightPosition[3] := 1;
ShadowSize       := 512;
DC := GetDC(pHandle);
RC := CreateRenderingContext(DC, [opDoubleBuffered], 32, 24, 0, 0, 0, 0, DummyPal);
wglMakeCurrent(DC, RC);
glClearColor(0.2,0.2,0.2,1);
glEnable(GL_DEPTH_TEST);
glDepthFunc(GL_LESS);
// Lichteigenschaften
glEnable(GL_LIGHT0);
glLightfv(GL_LIGHT0, GL_DIFFUSE, @LightDiffuse);
glLightfv(GL_LIGHT0, GL_LINEAR_ATTENUATION , @LightLinearAtt);
glLightModelfv(GL_LIGHT_MODEL_AMBIENT, @LightAmbient);
CreateDepthTexture;
end;

procedure TShadowMapDemo.DeInitOpenGL;
begin
wglDeleteContext(RC);
DeleteDC(DC);
end;

// =============================================================================
// Erstellt die Tiefentextur
// =============================================================================
procedure TShadowMapDemo.CreateDepthTexture;
begin
if DepthTexture > 0 then
 glDeleteTextures(1, @DepthTexture);
// Tiefentextur erstellen
glGenTextures(1, @DepthTexture);
glBindTexture(GL_TEXTURE_2D, DepthTexture);
// GL_DEPTH_COMPONENT als internes Texturenformat verwenden.Mittels dieser Konstante wird die
// Przision des Formates an den Tiefenpuffer des Renderkontextes angepasst
glCopyTexImage2D(GL_TEXTURE_2D, 0, GL_DEPTH_COMPONENT, 0, 0, ShadowSize, ShadowSize, 0);
// Filtermethode GL_LINEAR, um die Schattentextur weicher erscheinen zu lassen
glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
// Tiefentextur fr den Z-Vergleich vorbereiten.Ihre Tiefenwerte werden mit der R-Texturenkoordinate
// verglichen, die den Tiefenwert des Fragments im Koordinatensystem des Lichtes reprsentiert.
// Wenn der Vergleich scheitert, liefert der Tet 0, ansonsten den Wert 1.
glTexParameteri(GL_TEXTURE_2D, GL_DEPTH_TEXTURE_MODE_ARB, GL_LUMINANCE);
glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_COMPARE_MODE_ARB, GL_COMPARE_R_TO_TEXTURE_ARB);
glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_COMPARE_FUNC_ARB, GL_LEQUAL);
end;

// =============================================================================
// Zeigt die Tiefentextur an
// =============================================================================
procedure TShadowMapDemo.DisplayDepthTexture;
begin
glViewport(0, 0, WindowSize.x, WindowSize.y);
glDisable(GL_LIGHTING);

glMatrixMode(GL_PROJECTION);
glLoadIdentity;
gluOrtho2D(-1.0, 1.0, -1.0, 1.0);
glMatrixMode(GL_MODELVIEW);
glLoadIdentity;
// Texturenmatrix zurcksetzen
glMatrixMode(GL_TEXTURE);
glLoadIdentity;
// Tiefentextur binden
glEnable(GL_TEXTURE_2D);
glBindTexture(GL_TEXTURE_2D, DepthTexture);
// Hardware Shadowmapping deaktivieren damit die Schattentextur korrekt dargestellt wird
glTexParameteri(GL_TEXTURE_2D, GL_DEPTH_TEXTURE_MODE_ARB, GL_FALSE);
glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_COMPARE_MODE_ARB, GL_FALSE);
glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_COMPARE_FUNC_ARB, GL_FALSE);
// PBuffer binden
// Tiefentextur anzeigen
glBegin(GL_QUADS);
 glTexCoord2f(0, 0);  glVertex3f(-1, -1, 0);
 glTexCoord2f(0, 1);  glVertex3f(-1,  1, 0);
 glTexCoord2f(1, 1);  glVertex3f( 1,  1, 0);
 glTexCoord2f(1, 0);  glVertex3f( 1, -1, 0);
glEnd;
// Hardware Shadowmapping wieder aktivieren
glTexParameteri(GL_TEXTURE_2D, GL_DEPTH_TEXTURE_MODE_ARB, GL_LUMINANCE);
glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_COMPARE_MODE_ARB, GL_COMPARE_R_TO_TEXTURE_ARB);
glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_COMPARE_FUNC_ARB, GL_LEQUAL);
end;

// =============================================================================
// Erstellt eine Textur mit Tiefeninformationen der Szene
// Aufruf ist eigentlich nur ntig, wenn sich Objekte innerhalb der Szene
// bewegen, da Schatten vom Blickwinkel unabhngig sind
// Allerdings wird die Textur in diesem Demo aufgrund der Faulheit des Autors
// in jedem Frame erstellt ;-)
// =============================================================================
procedure TShadowMapDemo.UpdateDepthTexture;
const
 OriginPosition : array[0..3] of Single = (0, 0, 0, 1);
begin
glViewPort(0,0,ShadowSize,ShadowSize);
glMatrixMode(GL_PROJECTION);
glPushMatrix;
glLoadIdentity;
gluPerspective(90, WindowSize.x/WindowSize.y, ClipNear, ClipFar);
glMatrixMode(GL_MODELVIEW);
glClear(GL_COLOR_BUFFER_BIT or GL_DEPTH_BUFFER_BIT);
glLoadIdentity;
glDisable(GL_LIGHTING);
glColorMask(False, False, False, False);
// glPolygonOffset wird bentigt, da sonst auch Stellen schattiert werden, die davon verschont bleiben sollen.Warum das so ist
// erfhrt man in den entsprechenden Dokumenten von nVidia.Die beiden Werte (2,4) muss man durch probieren herausfinden, da
// sie je nach Szene variieren. 2 und 4 sind jedoch ein von nVidia vorgeschlagener, und meist ausreichender Wert
glPolygonOffset(2, 4);
glEnable(GL_POLYGON_OFFSET_FILL);
// Szene aus der Sicht des Lichtes rendern
glMultMatrixf(@LightLookAtMatrix);
// Backfaceculling aktivieren (kann je nach Geometrie jedoch Probleme verursachen)
glEnable(GL_CULL_FACE);
RenderScene(False);
glDisable(GL_CULL_FACE);

// Den Inhalt des Tiefenpuffers in die Schattentextur kopieren
glBindTexture(GL_TEXTURE_2D, DepthTexture);
glCopyTexSubImage2D(GL_TEXTURE_2D, 0, 0, 0, 0, 0, ShadowSize, ShadowSize);

glDisable(GL_POLYGON_OFFSET_FILL);
glColorMask(True, True, True, True);

glMatrixMode(GL_PROJECTION);
glPopMatrix;
glViewPort(0,0,WindowSize.x,WindowSize.y);
glMatrixMode(GL_MODELVIEW);
end;

// =============================================================================
// Zeichnet die Szene mit verschiedenen Objektkonfigurationen
// =============================================================================
procedure TShadowMapDemo.RenderScene(pDrawFloor : Boolean);
var
 i,j,k : Integer;
begin
if WireFrame then
 glPolygonMode(GL_FRONT_AND_BACK, GL_LINE)
else
 glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
glMatrixMode(GL_MODELVIEW);
case ObjConfig of
 0 : for i := 0 to 2 do
      for j := 0 to 2 do
       for k := 0 to 2 do
        begin
        if (i+j+k) mod 2 = 0 then
         glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, @ColorRed)
        else
         glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, @ColorGreen);
        glPushMatrix;
         glTranslatef(-1+i, 1+j, -1+k);
         gloSolidSphere(0.4, 16, 8);
        glPopMatrix;
        end;
 1 : begin
     glPushMatrix;
      glTranslatef(0, 2, 0);
      gloSolidSphere(0.6, 32, 32);
      if ObjRotate then
       glRotatef(Spin*4, 0, 1, 0);
      glTranslatef(-1, -1, 0);
      gloSolidSphere(0.2, 12, 12);
     glPopMatrix;
     end;
 2 : begin
     glPushMatrix;
      glTranslatef(0, 2, 0);
      glRotatef(90, 1, 0, 0);
      glRotatef(-45, 0, 1, 0);
      if ObjRotate then
       glRotatef(Spin, 0, 1, 0);
      gloSolidTorus(0.5, 1, 32, 32);
     glPopMatrix;
     end;
 3 : begin
     glPushMatrix;
      glTranslatef(0, 2.5, 0);
      if ObjRotate then
       glRotatef(Spin, 0, 1, 0);
      gloSolidTeapot(1);
     glPopMatrix;
     glPushMatrix;
      glScalef(1, 0.75, 1);
      gloSolidCube(3);
     glPopMatrix;
     end;
 4 : begin
     glPushMatrix;
      glTranslatef(0, 2, 0);
      glRotatef(90, 1, 0, 0);
      glRotatef(-45, 0, 1, 0);
      if ObjRotate then
       glRotatef(Spin, 0, 0, 1);
      gloSolidGear(0.5, 1, 0.2, 32, 0.2);
     glPopMatrix;
     glPushMatrix;
      glTranslatef(0, 1, 0);
      glRotatef(90, 1, 0, 0);
      glRotatef(-45, 0, 1, 0);
      gloSolidGear(0.3, 0.7, 0.2, 32, 0.1);
     glPopMatrix;
     end;
end;
if pDrawFloor then
 begin
 glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, @ColorGray);
 glBegin( GL_QUADS );
  for i := -2 to 11 do
   for j := -2 to 11 do
    begin
    glNormal3f(0, 1, 0);
    glVertex3f(-5+i,   0, -5+j);
    glVertex3f(-5+i,   0, -5+j+1);
    glVertex3f(-5+i+1, 0, -5+j+1);
    glVertex3f(-5+i+1, 0, -5+j);
    end;
 glEnd;
 end;
end;

// =============================================================================
// Rendert die Szene mit der projezierten Tiefentextur
// =============================================================================
procedure TShadowMapDemo.Render;
const
 x : array[0..3] of Single = (1, 0, 0, 0);
 y : array[0..3] of Single = (0, 1, 0, 0);
 z : array[0..3] of Single = (0, 0, 1, 0);
 w : array[0..3] of Single = (0, 0, 0, 1);
begin
glMatrixMode(GL_PROJECTION);
glLoadIdentity;
gluPerspective(45, WindowSize.x/WindowSize.y, ClipNear, ClipFar);
// Hier erstellen wir eine Matrix, die den Blickwinkel des Lichtes reprsentiert und speichern diese fr eine
// sptere Nutzung
glMatrixMode(GL_MODELVIEW);
glLoadIdentity;
gluLookAt(LightPosition[0],LightPosition[1],LightPosition[2], 0,2.5,0,  0,1,0);
// Modelansichtsmatrix auslesen und im Array ab legen
glGetFloatv(GL_MODELVIEW_MATRIX, @LightLookAtMatrix);
// Tiefentextur erstellen
UpdateDepthTexture;
glClear(GL_COLOR_BUFFER_BIT or GL_DEPTH_BUFFER_BIT);
// Betrachteransicht setzen
glMatrixMode(GL_MODELVIEW);
glLoadIdentity;
glTranslatef(0, -2, -15);
glRotatef(-SpinY_L, 1, 0, 0);
glRotatef(-SpinX_L, 0, 1, 0);
// Die Lichtposition als gelbe Kugel rendern
glDisable(GL_LIGHTING);
glPushMatrix;
 // Lichtquelle positionieren
 glLightfv(GL_LIGHT0, GL_POSITION, @LightPosition);
 glTranslatef(LightPosition[0], LightPosition[1], LightPosition[2]);
 glColor3f(1, 1, 0.5);
 gloSolidSphere(0.1, 8, 8);
 glColor3f(1, 1, 1);
glPopMatrix;
glEnable(GL_LIGHTING);
// Die folgenden Zeilen (glTexGenfv, glTexGeni, glEnable) bereiten OpenGLs Statemachine fr das projektive Texturemapping der
// Tiefentextur vor
glEnable(GL_LIGHTING);

glEnable(GL_TEXTURE_GEN_S);
glEnable(GL_TEXTURE_GEN_T);
glEnable(GL_TEXTURE_GEN_R);
glEnable(GL_TEXTURE_GEN_Q);

glTexGenfv(GL_S, GL_EYE_PLANE, @x);
glTexGenfv(GL_T, GL_EYE_PLANE, @y);
glTexGenfv(GL_R, GL_EYE_PLANE, @z);
glTexGenfv(GL_Q, GL_EYE_PLANE, @w);

glTexGeni(GL_S, GL_TEXTURE_GEN_MODE, GL_EYE_LINEAR);
glTexGeni(GL_T, GL_TEXTURE_GEN_MODE, GL_EYE_LINEAR);
glTexGeni(GL_R, GL_TEXTURE_GEN_MODE, GL_EYE_LINEAR);
glTexGeni(GL_Q, GL_TEXTURE_GEN_MODE, GL_EYE_LINEAR);

// In die Texturenmatrix wechseln und das projektive Texturenmapping vorbereiten
glMatrixMode(GL_TEXTURE);
glLoadIdentity;

glTranslatef(0.5, 0.5, 0.5);
glScalef(0.5, 0.5, 0.5);
gluPerspective(90, WindowSize.x/WindowSize.y, ClipNear, ClipFar);

glMultMatrixf(@LightLookAtMatrix);
// Tiefentextur binden damit sie als Schattentextur projeziert wird
glEnable(GL_TEXTURE_2D);
glBindTexture(GL_TEXTURE_2D, DepthTexture);

RenderScene(True);

glDisable(GL_TEXTURE_2D);
glDisable(GL_TEXTURE_GEN_S);
glDisable(GL_TEXTURE_GEN_T);
glDisable(GL_TEXTURE_GEN_R);
glDisable(GL_TEXTURE_GEN_Q);

if RenderDepthTex then
 DisplayDepthTexture;

SwapBuffers(DC);
end;

end.

