// =============================================================================
//
//   glFont.pas
//
// =============================================================================
//   Copyright  2003-2004 by Sascha Willems - http://www.delphigl.de
//   visit the Delphi OpenGL Community - http://www.delphigl.com
// =============================================================================
//   Contents of this file are subject to the GNU Public License (GPL) which can
//   be obtained here : http://opensource.org/licenses/gpl-license.php
//   So only use this file if you fully unterstand that license!!!
// =============================================================================
//   Unit for drawing fonts in opengl
// =============================================================================

unit glFont;

interface

uses
 Windows,
 SysUtils,
 dglOpenGL,
 Textures;

type
 TTexFont = class(TObject)
   FontWidth   : array[0..255] of Byte;
   FontBase    : Cardinal;
   FontTexture : glUInt;
   constructor Create(pFontTextureFile : String); overload;
   destructor Destroy; override;
   procedure Print3D(pText : String;pPosition : array of Single;pScaleFactor : Single);
   procedure Print2D(pText : String;pPosition,pColor : array of Single;pAlpha, pScaleFactor : Single);
   procedure Print2DEx(pText : String;pPosition,pColor : array of Single;pAlpha, pScaleFactor : Single;pCenter : Boolean);
  end;

var
 TextureFont  : TTexFont;

implementation

// =============================================================================
// =============================================================================
//  TTexFont
// =============================================================================
// =============================================================================

// =============================================================================
//  TTexFont.Create                                              - CONSTRUCTOR -
// =============================================================================
constructor TTexFont.Create(pFontTextureFile : String);
var
 F      : File of Byte;
 i      : Byte;
 x,y    : Single;
 xs     : Single;
begin
inherited Create;
LoadTexture(pFontTextureFile, FontTexture);
if not FileExists('font.fnt') then
 begin
 for i := 0 to 255 do
  FontWidth[i] := 16;
 end
else
 begin
 AssignFile(F, 'font.fnt');
 Reset(F);
 for i := 0 to 255 do
  Read(F, FontWidth[i]);
 CloseFile(F);
 end;
FontBase := glGenLists(256);
for i := 0 to 255 do
 begin
 x := (i mod 16) / 16;
 y := (i div 16) / 16;
 glNewList(FontBase+i, GL_COMPILE);
  glBegin(GL_QUADS);
   xs := (16-FontWidth[I])/2/256;
   glTexCoord2f(x+xs,                            1-y);        glVertex2i(0, 0);
   glTexCoord2f(x+xs + FontWidth[I]/256 + 1/512, 1-y);        glVertex2i(FontWidth[I], 0);
   glTexCoord2f(x+xs + FontWidth[I]/256 + 1/512, 1-y-0.0625); glVertex2i(FontWidth[I], 16);
   glTexCoord2f(x+xs,                            1-y-0.0625); glVertex2i(0, 16);
  glEnd;
  glTranslatef(FontWidth[i]+1, 0, 0);
 glEndList;
 end;
end;

// =============================================================================
//  TTexFont.Destroy                                              - DESTRUCTOR -
// =============================================================================
destructor TTexFont.Destroy;
begin
inherited;
glDeleteLists(FontBase, 256);
end;

// =============================================================================
//  TTexFont.Print3D
// =============================================================================
procedure TTexFont.Print3D(pText : String;pPosition : array of Single;pScaleFactor : Single);
begin
glPushMatrix;
 glTranslatef(pPosition[0],pPosition[1],pPosition[2]);
 glScalef(pScaleFactor, pScaleFactor , pScaleFactor);
 glListBase(FontBase);
 glCallLists(Length(pText), GL_BYTE, Pointer(pText));
glPopMatrix;
end;

// =============================================================================
//  TTexFont.Print2D
// =============================================================================
procedure TTexFont.Print2D(pText : String;pPosition,pColor : array of Single;pAlpha, pScaleFactor  : Single);
begin
if Length(pText) = 0 then
 exit;
glBindTexture(GL_TEXTURE_2D, FontTexture);
if pAlpha < 1 then
 begin
 glEnable(GL_BLEND);
 glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
 end;
glColor4f(pColor[0], pColor[1], pColor[2], pAlpha);
glPushMatrix;
 glTranslatef(pPosition[0],pPosition[1],pPosition[2]);
 glScalef(pScaleFactor, pScaleFactor , pScaleFactor);
 glListBase(FontBase);
 glCallLists(Length(pText), GL_BYTE, PChar(pText));
glPopMatrix;
glColor4f(1,1,1,1);
glDisable(GL_BLEND);
end;

// =============================================================================
//  TTexFont.Print2DEx
// =============================================================================
procedure TTexFont.Print2DEx(pText : String;pPosition,pColor : array of Single;pAlpha, pScaleFactor : Single;pCenter : Boolean);
begin
glDisable(GL_DEPTH_TEST);
glBindTexture(GL_TEXTURE_2D, FontTexture);
if pAlpha < 1 then
 begin
 glEnable(GL_BLEND);
 glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
 end;
glColor4f(pColor[0], pColor[1], pColor[2], pAlpha);
glPushMatrix;
 if pCenter then
  pPosition[0] := pPosition[0]-Length(pText)*6*pScaleFactor;
 glTranslatef(pPosition[0],pPosition[1],pPosition[2]);
 glScalef(pScaleFactor, pScaleFactor , pScaleFactor);
 glListBase(FontBase);
 glCallLists(Length(pText), GL_BYTE, Pointer(pText));
glPopMatrix;
glColor4f(1,1,1,1);
glEnable(GL_DEPTH_TEST);
glDisable(GL_BLEND);
end;

end.

